// ############################################################################################
//  SkyObjectFractals.cc
//  Implementation of various CIndefiniteAreaFractal derivative classes
//  specialized to imitate sky objects
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the algorithms used in this file were written based on the book :
//    Fractal Vision : Put Fractals to Work for You
//    Dick Oliver
//    Sams Publishing
//    1992
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "SkyObjectFractals.h"


namespace vge {

//
// Cirrus 1
//
FractalCirrus::FractalCirrus()
{
  if(SEED_REPLICATOR_FRACTAL_ALLOC(1, 6, 4)) {
    SEED_POINT(0, 0,   75,  50);
    SEED_POINT(1, 0,  100,   0);
    SEED_POINT(2, 0,   75, -50);
    SEED_POINT(3, 0,  -75, -50);
    SEED_POINT(4, 0, -100,   0);
    SEED_POINT(5, 0,  -75,  50);
    SEED_REPLICA(0, 45, 45, -20, -20, -90, -12); // Left
    SEED_REPLICA(1, 40, 40,  00,  00,   0,  10); // Middle
    SEED_REPLICA(2, 80, 20,   0,   0,   0, -40); // Bottom
    SEED_REPLICA(3, 70, 45, 200, 200, 135,   0); // Right
  }
}

bool FractalCirrus::CosmosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_4(un32TotPart, un32CurPart, un32TotRelPart, un32CurRelPart);
  uint32_t un32C = uint32_t((31.0f / un32MaxLevel) * un32CurLevel);
  SetColor(uint8_t(un32C * 0.85f), uint8_t(un32C * 0.95f), un32C);
  SetColor(uint8_t(un32C * 0.85f), uint8_t(un32C * 0.95f), un32C);
  return(true);
}

bool FractalCirrus::ChaosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_4(un32MaxLevel, un32CurLevel, un32TotRelPart, un32CurRelPart);
  uint32_t un32C = uint32_t(_rnFloat.GeneratePositive(3)) + 28;
  SetColor(uint8_t(un32C * 0.85f), uint8_t(un32C * 0.95f), un32C);
  return(true);
}


//
// Stargate 1
//
FractalStargate::FractalStargate()
{
  if(SEED_REPLICATOR_FRACTAL_ALLOC(1, 16, 6)) {
    SEED_POINT(0,  0,    10,  60);
    SEED_POINT(0,  1,    50,  30);
    SEED_POINT(0,  2,    80,  80);
    SEED_POINT(0,  3,    80,  30);
    SEED_POINT(0,  4,    70, -10);
    SEED_POINT(0,  5,    30, -40);
    SEED_POINT(0,  6,    80, -70);
    SEED_POINT(0,  7,    30, -70);
    SEED_POINT(0,  8,   -10, -60);
    SEED_POINT(0,  9,   -50, -30);
    SEED_POINT(0, 10,   -80, -80);
    SEED_POINT(0, 11,   -80, -30);
    SEED_POINT(0, 12,   -70,  10);
    SEED_POINT(0, 13,   -30,  40);
    SEED_POINT(0, 14,   -80,  70);
    SEED_POINT(0, 15,   -30,  70);
    SEED_REPLICA(0, 75, 75, -256, -256,   0,   0, false); // Middle #1
    SEED_REPLICA(1, 55, 55, -768, -768,   0,   0, false); // Middle #1
    SEED_REPLICA(2, 25, 25,    0,    0, -80,  70); // Top-Left
    SEED_REPLICA(3, 25, 25, 1024, 1024, -80, -80); // Bottom-Left
    SEED_REPLICA(4, 25, 25, 2048, 2048,  80, -70); // Bottom-Right
    SEED_REPLICA(5, 25, 25, 3072, 3072,  80,  80); // Top-Right
  }
}

bool FractalStargate::CosmosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_4(un32TotPart, un32CurPart, un32TotRelPart, un32CurRelPart);
  float    fFac  = float(un32MaxLevel - un32CurLevel) / float(un32MaxLevel);
  float    fMax  = sqrt(fFac) * 31.0f;
  float    fMin  = fFac * 31.0f;
  float    fDel  = fMax - fMin;
  uint32_t un32C = uint32_t(_rnFloat.GeneratePositive(fDel) + fMin);
  SetColor(uint8_t(un32C * 0.85f), uint8_t(un32C * 0.95f), un32C);
  return(true);
}

bool FractalStargate::ChaosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_2(un32TotRelPart, un32CurRelPart);
  float    fFac  = float(un32CurLevel) / float(un32MaxLevel);
  float    fMax  = fFac * 31.0f;
  float    fMin  = fFac * fFac * 31.0f;
  float    fDel  = fMax - fMin;
  uint32_t un32C = uint32_t(_rnFloat.GeneratePositive(fDel) + fMin);
  SetColor(uint8_t(un32C * 0.85f), uint8_t(un32C * 0.95f), un32C);
  return(true);
}

} // namespace vge
