// ############################################################################################
//  Matrix3D.cc
//  Implementation of CMatrix3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/LookUpTables.h"
#include "../Base/MessageDisplayer.h"
#include "Matrix3D.h"


namespace vge {

//
// The 3D matrix class
//
void CMatrix3D::DumpMatrix()
{
  VInfo(_NFL_, "|   %+10.3f   %+10.3f   %+10.3f   %+10.3f   |\n|   %+10.3f   %+10.3f   %+10.3f   %+10.3f   |\n|   %+10.3f   %+10.3f   %+10.3f   %+10.3f   |\n|   %+10.3f   %+10.3f   %+10.3f   %+10.3f   |",
    _aafMatrix[0][0], _aafMatrix[0][1], _aafMatrix[0][2], _aafMatrix[0][3], _aafMatrix[1][0], _aafMatrix[1][1], _aafMatrix[1][2], _aafMatrix[1][3],
    _aafMatrix[2][0], _aafMatrix[2][1], _aafMatrix[2][2], _aafMatrix[2][3], _aafMatrix[3][0], _aafMatrix[3][1], _aafMatrix[3][2], _aafMatrix[3][3]);
}

void CMatrix3D::SetToLookAtMatrix(const float afCamPosXYZ[3], const float afCamTargetXYZ[3])
{
  // Static variables
  static float afDefUpVect[3] = {0, 1, 0};

  // Create matrix
  SetToLookAtMatrix(afCamPosXYZ, afCamTargetXYZ, afDefUpVect);
}

void CMatrix3D::SetToLookAtMatrix(const float afCamPosXYZ[3], const float afCamTargetXYZ[3], const float afUpVectXYZ[3])
{
  // Temporary variables
  float afX[3], afY[3], afZ[3], fMagX, fMagY, fMagZ, fDotX, fDotY, fDotZ;

  // Calculate the z-axis
  afZ[0] = afCamTargetXYZ[0] - afCamPosXYZ[0];
  afZ[1] = afCamTargetXYZ[1] - afCamPosXYZ[1];
  afZ[2] = afCamTargetXYZ[2] - afCamPosXYZ[2];
  fMagZ  = sqrt(afZ[0] * afZ[0] + afZ[1] * afZ[1] + afZ[2] * afZ[2]);
  afZ[0] /= fMagZ; afZ[1] /= fMagZ; afZ[2] /= fMagZ;

  // Calculate the x-axis
  afX[0] = afUpVectXYZ[1] * afZ[2] - afUpVectXYZ[2] * afZ[1];
  afX[1] = afUpVectXYZ[2] * afZ[0] - afUpVectXYZ[0] * afZ[2];
  afX[2] = afUpVectXYZ[0] * afZ[1] - afUpVectXYZ[1] * afZ[0];
  fMagX  = sqrt(afX[0] * afX[0] + afX[1] * afX[1] + afX[2] * afX[2]);
  afX[0] /= fMagX; afX[1] /= fMagX; afX[2] /= fMagX;

  // Calculate the y-axis
  afY[0] = afZ[1] * afX[2] - afZ[2] * afX[1];
  afY[1] = afZ[2] * afX[0] - afZ[0] * afX[2];
  afY[2] = afZ[0] * afX[1] - afZ[1] * afX[0];
  fMagY  = sqrt(afY[0] * afY[0] + afY[1] * afY[1] + afY[2] * afY[2]);

  // Calculate the dot product
  fDotX = afX[0] * afCamPosXYZ[0] + afX[1] * afCamPosXYZ[1] + afX[2] * afCamPosXYZ[2];
  fDotY = afY[0] * afCamPosXYZ[0] + afY[1] * afCamPosXYZ[1] + afY[2] * afCamPosXYZ[2];
  fDotZ = afZ[0] * afCamPosXYZ[0] + afZ[1] * afCamPosXYZ[1] + afZ[2] * afCamPosXYZ[2];

  // Set matrix
  _aafMatrix[0][0] =  afX[0]; _aafMatrix[0][1] =  afY[0]; _aafMatrix[0][2] =  afZ[0]; _aafMatrix[0][3] = 0;
  _aafMatrix[1][0] =  afX[1]; _aafMatrix[1][1] =  afY[1]; _aafMatrix[1][2] =  afZ[1]; _aafMatrix[1][3] = 0;
  _aafMatrix[2][0] =  afX[2]; _aafMatrix[2][1] =  afY[2]; _aafMatrix[2][2] =  afZ[2]; _aafMatrix[2][3] = 0;
  _aafMatrix[3][0] = -fDotX;  _aafMatrix[3][1] = -fDotY;  _aafMatrix[3][2] = -fDotZ;  _aafMatrix[3][3] = 0;
}

void CMatrix3D::SetToShearingMatrix(float fX, float fY)
{
  // Create shearing matrix
  _aafMatrix[0][0] = 1; _aafMatrix[0][1] = 0; _aafMatrix[0][2] = fX; _aafMatrix[0][3] = 0;
  _aafMatrix[1][0] = 0; _aafMatrix[1][1] = 1; _aafMatrix[1][2] = fY; _aafMatrix[1][3] = 0;
  _aafMatrix[2][0] = 0; _aafMatrix[2][1] = 0; _aafMatrix[2][2] = 1;  _aafMatrix[2][3] = 0;
  _aafMatrix[3][0] = 0; _aafMatrix[3][1] = 0; _aafMatrix[3][2] = 0;  _aafMatrix[3][3] = 1;
}

void CMatrix3D::SetToScallingMatrix(float fX, float fY, float fZ)
{
  // Create scalling matrix
  _aafMatrix[0][0] = fX; _aafMatrix[0][1] = 0;  _aafMatrix[0][2] = 0;  _aafMatrix[0][3] = 0;
  _aafMatrix[1][0] = 0;  _aafMatrix[1][1] = fY; _aafMatrix[1][2] = 0;  _aafMatrix[1][3] = 0;
  _aafMatrix[2][0] = 0;  _aafMatrix[2][1] = 0;  _aafMatrix[2][2] = fZ; _aafMatrix[2][3] = 0;
  _aafMatrix[3][0] = 0;  _aafMatrix[3][1] = 0;  _aafMatrix[3][2] = 0;  _aafMatrix[3][3] = 1;
}

void CMatrix3D::SetToRotatationMatrix(int32_t n32X, int32_t n32Y, int32_t n32Z)
{
  // Temporary variables
  float _sse aafMat[4][4];

  // Create identity matrix
  MakeIdentityMatrix(_aafMatrix);

  // Create Z rotation matrix
  aafMat[0][0] =  COS(n32Z); aafMat[0][1] = SIN(n32Z); aafMat[0][2] = 0; aafMat[0][3] = 0;
  aafMat[1][0] = -SIN(n32Z); aafMat[1][1] = COS(n32Z); aafMat[1][2] = 0; aafMat[1][3] = 0;
  aafMat[2][0] = 0;          aafMat[2][1] = 0;         aafMat[2][2] = 1; aafMat[2][3] = 0;
  aafMat[3][0] = 0;          aafMat[3][1] = 0;         aafMat[3][2] = 0; aafMat[3][3] = 1;
  _MergeMatrices_(_aafMatrix, aafMat);

  // Create X rotation matrix
  aafMat[0][0] = 1; aafMat[0][1] = 0;          aafMat[0][2] = 0;         aafMat[0][3] = 0;
  aafMat[1][0] = 0; aafMat[1][1] =  COS(n32X); aafMat[1][2] = SIN(n32X); aafMat[1][3] = 0;
  aafMat[2][0] = 0; aafMat[2][1] = -SIN(n32X); aafMat[2][2] = COS(n32X); aafMat[2][3] = 0;
  aafMat[3][0] = 0; aafMat[3][1] = 0;          aafMat[3][2] = 0;         aafMat[3][3] = 1;
  _MergeMatrices_(_aafMatrix, aafMat);

  // Create Y rotation matrix
  aafMat[0][0] = COS(n32Y); aafMat[0][1] = 0; aafMat[0][2] = -SIN(n32Y); aafMat[0][3] = 0;
  aafMat[1][0] = 0;         aafMat[1][1] = 1; aafMat[1][2] = 0;          aafMat[1][3] = 0;
  aafMat[2][0] = SIN(n32Y); aafMat[2][1] = 0; aafMat[2][2] = COS(n32Y);  aafMat[2][3] = 0;
  aafMat[3][0] = 0;         aafMat[3][1] = 0; aafMat[3][2] = 0;          aafMat[3][3] = 1;
  _MergeMatrices_(_aafMatrix, aafMat);
}

void CMatrix3D::SetToTranslationMatrix(float fX, float fY, float fZ)
{
  // Create translation matrix
  _aafMatrix[0][0] = 1;  _aafMatrix[0][1] = 0;  _aafMatrix[0][2] = 0;  _aafMatrix[0][3] = 0;
  _aafMatrix[1][0] = 0;  _aafMatrix[1][1] = 1;  _aafMatrix[1][2] = 0;  _aafMatrix[1][3] = 0;
  _aafMatrix[2][0] = 0;  _aafMatrix[2][1] = 0;  _aafMatrix[2][2] = 1;  _aafMatrix[2][3] = 0;
  _aafMatrix[3][0] = fX; _aafMatrix[3][1] = fY; _aafMatrix[3][2] = fZ; _aafMatrix[3][3] = 1;
}

// Matrix inversion
void CMatrix3D::InvertMatrices(float aafDstMatrix[4][4], const float aafSrcMatrix[4][4]) const
{
  // Temporary variables
  float afTmp[12];
  float aafTSrc[4][4];
  float fDet;

  // Transpose the source matrix
  aafTSrc[0][0] = aafSrcMatrix[0][0]; aafTSrc[0][1] = aafSrcMatrix[1][0]; aafTSrc[0][2] = aafSrcMatrix[2][0]; aafTSrc[0][3] = aafSrcMatrix[3][0];
  aafTSrc[1][0] = aafSrcMatrix[0][1]; aafTSrc[1][1] = aafSrcMatrix[1][1]; aafTSrc[1][2] = aafSrcMatrix[2][1]; aafTSrc[1][3] = aafSrcMatrix[3][1];
  aafTSrc[2][0] = aafSrcMatrix[0][2]; aafTSrc[2][1] = aafSrcMatrix[1][2]; aafTSrc[2][2] = aafSrcMatrix[2][2]; aafTSrc[2][3] = aafSrcMatrix[3][2];
  aafTSrc[3][0] = aafSrcMatrix[0][3]; aafTSrc[3][1] = aafSrcMatrix[1][3]; aafTSrc[3][2] = aafSrcMatrix[2][3]; aafTSrc[3][3] = aafSrcMatrix[3][3];

  // Calculate pairs for the first 8 cofactors
  afTmp[0]  = aafTSrc[2][2] * aafTSrc[3][3]; afTmp[1]  = aafTSrc[2][3] * aafTSrc[3][2];
  afTmp[2]  = aafTSrc[2][1] * aafTSrc[3][3]; afTmp[3]  = aafTSrc[2][3] * aafTSrc[3][1];
  afTmp[4]  = aafTSrc[2][1] * aafTSrc[3][2]; afTmp[5]  = aafTSrc[2][2] * aafTSrc[3][1];
  afTmp[6]  = aafTSrc[2][0] * aafTSrc[3][3]; afTmp[7]  = aafTSrc[2][3] * aafTSrc[3][0];
  afTmp[8]  = aafTSrc[2][0] * aafTSrc[3][2]; afTmp[9]  = aafTSrc[2][2] * aafTSrc[3][0];
  afTmp[10] = aafTSrc[2][0] * aafTSrc[3][1]; afTmp[11] = aafTSrc[2][1] * aafTSrc[3][0];

  // Calculate the first 8 cofactors
  aafDstMatrix[0][0] = (afTmp[0] * aafTSrc[1][1] + afTmp[3] * aafTSrc[1][2] + afTmp[4]  * aafTSrc[1][3]) - (afTmp[1] * aafTSrc[1][1] + afTmp[2] * aafTSrc[1][2] + afTmp[5]  * aafTSrc[1][3]);
  aafDstMatrix[0][1] = (afTmp[1] * aafTSrc[1][0] + afTmp[6] * aafTSrc[1][2] + afTmp[9]  * aafTSrc[1][3]) - (afTmp[0] * aafTSrc[1][0] + afTmp[7] * aafTSrc[1][2] + afTmp[8]  * aafTSrc[1][3]);
  aafDstMatrix[0][2] = (afTmp[2] * aafTSrc[1][0] + afTmp[7] * aafTSrc[1][1] + afTmp[10] * aafTSrc[1][3]) - (afTmp[3] * aafTSrc[1][0] + afTmp[6] * aafTSrc[1][1] + afTmp[11] * aafTSrc[1][3]);
  aafDstMatrix[0][3] = (afTmp[5] * aafTSrc[1][0] + afTmp[8] * aafTSrc[1][1] + afTmp[11] * aafTSrc[1][2]) - (afTmp[4] * aafTSrc[1][0] + afTmp[9] * aafTSrc[1][1] + afTmp[10] * aafTSrc[1][2]);
  aafDstMatrix[1][0] = (afTmp[1] * aafTSrc[0][1] + afTmp[2] * aafTSrc[0][2] + afTmp[5]  * aafTSrc[0][3]) - (afTmp[0] * aafTSrc[0][1] + afTmp[3] * aafTSrc[0][2] + afTmp[4]  * aafTSrc[0][3]);
  aafDstMatrix[1][1] = (afTmp[0] * aafTSrc[0][0] + afTmp[7] * aafTSrc[0][2] + afTmp[8]  * aafTSrc[0][3]) - (afTmp[1] * aafTSrc[0][0] + afTmp[6] * aafTSrc[0][2] + afTmp[9]  * aafTSrc[0][3]);
  aafDstMatrix[1][2] = (afTmp[3] * aafTSrc[0][0] + afTmp[6] * aafTSrc[0][1] + afTmp[11] * aafTSrc[0][3]) - (afTmp[2] * aafTSrc[0][0] + afTmp[7] * aafTSrc[0][1] + afTmp[10] * aafTSrc[0][3]);
  aafDstMatrix[1][3] = (afTmp[4] * aafTSrc[0][0] + afTmp[9] * aafTSrc[0][1] + afTmp[10] * aafTSrc[0][2]) - (afTmp[5] * aafTSrc[0][0] + afTmp[8] * aafTSrc[0][1] + afTmp[11] * aafTSrc[0][2]);

  // Calculate pairs for the second 8 cofactors
  afTmp[0]  = aafTSrc[0][2] * aafTSrc[1][3]; afTmp[1]  = aafTSrc[0][3] * aafTSrc[1][2];
  afTmp[2]  = aafTSrc[0][1] * aafTSrc[1][3]; afTmp[3]  = aafTSrc[0][3] * aafTSrc[1][1];
  afTmp[4]  = aafTSrc[0][1] * aafTSrc[1][2]; afTmp[5]  = aafTSrc[0][2] * aafTSrc[1][1];
  afTmp[6]  = aafTSrc[0][0] * aafTSrc[1][3]; afTmp[7]  = aafTSrc[0][3] * aafTSrc[1][0];
  afTmp[8]  = aafTSrc[0][0] * aafTSrc[1][2]; afTmp[9]  = aafTSrc[0][2] * aafTSrc[1][0];
  afTmp[10] = aafTSrc[0][0] * aafTSrc[1][1]; afTmp[11] = aafTSrc[0][1] * aafTSrc[1][0];

  // Calculate the second 8 cofactors
  aafDstMatrix[2][0] = (afTmp[0]  * aafTSrc[3][1] + afTmp[3]  * aafTSrc[3][2] + afTmp[4]  * aafTSrc[3][3]) - (afTmp[1]  * aafTSrc[3][1] + afTmp[2]  * aafTSrc[3][2] + afTmp[5]  * aafTSrc[3][3]);
  aafDstMatrix[2][1] = (afTmp[1]  * aafTSrc[3][0] + afTmp[6]  * aafTSrc[3][2] + afTmp[9]  * aafTSrc[3][3]) - (afTmp[0]  * aafTSrc[3][0] + afTmp[7]  * aafTSrc[3][2] + afTmp[8]  * aafTSrc[3][3]);
  aafDstMatrix[2][2] = (afTmp[2]  * aafTSrc[3][0] + afTmp[7]  * aafTSrc[3][1] + afTmp[10] * aafTSrc[3][3]) - (afTmp[3]  * aafTSrc[3][0] + afTmp[6]  * aafTSrc[3][1] + afTmp[11] * aafTSrc[3][3]);
  aafDstMatrix[2][3] = (afTmp[5]  * aafTSrc[3][0] + afTmp[8]  * aafTSrc[3][1] + afTmp[11] * aafTSrc[3][2]) - (afTmp[4]  * aafTSrc[3][0] + afTmp[9]  * aafTSrc[3][1] + afTmp[10] * aafTSrc[3][2]);
  aafDstMatrix[3][0] = (afTmp[2]  * aafTSrc[2][2] + afTmp[5]  * aafTSrc[2][3] + afTmp[1]  * aafTSrc[2][1]) - (afTmp[4]  * aafTSrc[2][3] + afTmp[0]  * aafTSrc[2][1] + afTmp[3]  * aafTSrc[2][2]);
  aafDstMatrix[3][1] = (afTmp[8]  * aafTSrc[2][3] + afTmp[0]  * aafTSrc[2][0] + afTmp[7]  * aafTSrc[2][2]) - (afTmp[6]  * aafTSrc[2][2] + afTmp[9]  * aafTSrc[2][3] + afTmp[1]  * aafTSrc[2][0]);
  aafDstMatrix[3][2] = (afTmp[6]  * aafTSrc[2][1] + afTmp[11] * aafTSrc[2][3] + afTmp[3]  * aafTSrc[2][0]) - (afTmp[10] * aafTSrc[2][3] + afTmp[2]  * aafTSrc[2][0] + afTmp[7]  * aafTSrc[2][1]);
  aafDstMatrix[3][3] = (afTmp[10] * aafTSrc[2][2] + afTmp[4]  * aafTSrc[2][0] + afTmp[9]  * aafTSrc[2][1]) - (afTmp[8]  * aafTSrc[2][1] + afTmp[11] * aafTSrc[2][2] + afTmp[5]  * aafTSrc[2][0]);

  // Calculate the 1/D value
  fDet = aafTSrc[0][0] * aafDstMatrix[0][0] + aafTSrc[0][1] * aafDstMatrix[0][1] + aafTSrc[0][2] * aafDstMatrix[0][2] + aafTSrc[0][3] * aafDstMatrix[0][3];
  fDet = 1 / fDet;

  // Multiply all elements with the 1/D value
  aafDstMatrix[0][0] *= fDet; aafDstMatrix[0][1] *= fDet; aafDstMatrix[0][2] *= fDet; aafDstMatrix[0][3] *= fDet;
  aafDstMatrix[1][0] *= fDet; aafDstMatrix[1][1] *= fDet; aafDstMatrix[1][2] *= fDet; aafDstMatrix[1][3] *= fDet;
  aafDstMatrix[2][0] *= fDet; aafDstMatrix[2][1] *= fDet; aafDstMatrix[2][2] *= fDet; aafDstMatrix[2][3] *= fDet;
  aafDstMatrix[3][0] *= fDet; aafDstMatrix[3][1] *= fDet; aafDstMatrix[3][2] *= fDet; aafDstMatrix[3][3] *= fDet;
}

void CMatrix3D::InvertMatrix(float aafMatrix[4][4]) const
{
  // Temporary variables
  float afTmp[12];
  float aafTSrc[4][4], aafDst[4][4];
  float fDet;

  // Transpose the source matrix
  aafTSrc[0][0] = aafMatrix[0][0]; aafTSrc[0][1] = aafMatrix[1][0]; aafTSrc[0][2] = aafMatrix[2][0]; aafTSrc[0][3] = aafMatrix[3][0];
  aafTSrc[1][0] = aafMatrix[0][1]; aafTSrc[1][1] = aafMatrix[1][1]; aafTSrc[1][2] = aafMatrix[2][1]; aafTSrc[1][3] = aafMatrix[3][1];
  aafTSrc[2][0] = aafMatrix[0][2]; aafTSrc[2][1] = aafMatrix[1][2]; aafTSrc[2][2] = aafMatrix[2][2]; aafTSrc[2][3] = aafMatrix[3][2];
  aafTSrc[3][0] = aafMatrix[0][3]; aafTSrc[3][1] = aafMatrix[1][3]; aafTSrc[3][2] = aafMatrix[2][3]; aafTSrc[3][3] = aafMatrix[3][3];

  // Calculate pairs for the first 8 cofactors
  afTmp[0]  = aafTSrc[2][2] * aafTSrc[3][3]; afTmp[1]  = aafTSrc[2][3] * aafTSrc[3][2];
  afTmp[2]  = aafTSrc[2][1] * aafTSrc[3][3]; afTmp[3]  = aafTSrc[2][3] * aafTSrc[3][1];
  afTmp[4]  = aafTSrc[2][1] * aafTSrc[3][2]; afTmp[5]  = aafTSrc[2][2] * aafTSrc[3][1];
  afTmp[6]  = aafTSrc[2][0] * aafTSrc[3][3]; afTmp[7]  = aafTSrc[2][3] * aafTSrc[3][0];
  afTmp[8]  = aafTSrc[2][0] * aafTSrc[3][2]; afTmp[9]  = aafTSrc[2][2] * aafTSrc[3][0];
  afTmp[10] = aafTSrc[2][0] * aafTSrc[3][1]; afTmp[11] = aafTSrc[2][1] * aafTSrc[3][0];

  // Calculate the first 8 cofactors
  aafDst[0][0] = (afTmp[0] * aafTSrc[1][1] + afTmp[3] * aafTSrc[1][2] + afTmp[4]  * aafTSrc[1][3]) - (afTmp[1] * aafTSrc[1][1] + afTmp[2] * aafTSrc[1][2] + afTmp[5]  * aafTSrc[1][3]);
  aafDst[0][1] = (afTmp[1] * aafTSrc[1][0] + afTmp[6] * aafTSrc[1][2] + afTmp[9]  * aafTSrc[1][3]) - (afTmp[0] * aafTSrc[1][0] + afTmp[7] * aafTSrc[1][2] + afTmp[8]  * aafTSrc[1][3]);
  aafDst[0][2] = (afTmp[2] * aafTSrc[1][0] + afTmp[7] * aafTSrc[1][1] + afTmp[10] * aafTSrc[1][3]) - (afTmp[3] * aafTSrc[1][0] + afTmp[6] * aafTSrc[1][1] + afTmp[11] * aafTSrc[1][3]);
  aafDst[0][3] = (afTmp[5] * aafTSrc[1][0] + afTmp[8] * aafTSrc[1][1] + afTmp[11] * aafTSrc[1][2]) - (afTmp[4] * aafTSrc[1][0] + afTmp[9] * aafTSrc[1][1] + afTmp[10] * aafTSrc[1][2]);
  aafDst[1][0] = (afTmp[1] * aafTSrc[0][1] + afTmp[2] * aafTSrc[0][2] + afTmp[5]  * aafTSrc[0][3]) - (afTmp[0] * aafTSrc[0][1] + afTmp[3] * aafTSrc[0][2] + afTmp[4]  * aafTSrc[0][3]);
  aafDst[1][1] = (afTmp[0] * aafTSrc[0][0] + afTmp[7] * aafTSrc[0][2] + afTmp[8]  * aafTSrc[0][3]) - (afTmp[1] * aafTSrc[0][0] + afTmp[6] * aafTSrc[0][2] + afTmp[9]  * aafTSrc[0][3]);
  aafDst[1][2] = (afTmp[3] * aafTSrc[0][0] + afTmp[6] * aafTSrc[0][1] + afTmp[11] * aafTSrc[0][3]) - (afTmp[2] * aafTSrc[0][0] + afTmp[7] * aafTSrc[0][1] + afTmp[10] * aafTSrc[0][3]);
  aafDst[1][3] = (afTmp[4] * aafTSrc[0][0] + afTmp[9] * aafTSrc[0][1] + afTmp[10] * aafTSrc[0][2]) - (afTmp[5] * aafTSrc[0][0] + afTmp[8] * aafTSrc[0][1] + afTmp[11] * aafTSrc[0][2]);

  // Calculate pairs for the second 8 cofactors
  afTmp[0]  = aafTSrc[0][2] * aafTSrc[1][3]; afTmp[1]  = aafTSrc[0][3] * aafTSrc[1][2];
  afTmp[2]  = aafTSrc[0][1] * aafTSrc[1][3]; afTmp[3]  = aafTSrc[0][3] * aafTSrc[1][1];
  afTmp[4]  = aafTSrc[0][1] * aafTSrc[1][2]; afTmp[5]  = aafTSrc[0][2] * aafTSrc[1][1];
  afTmp[6]  = aafTSrc[0][0] * aafTSrc[1][3]; afTmp[7]  = aafTSrc[0][3] * aafTSrc[1][0];
  afTmp[8]  = aafTSrc[0][0] * aafTSrc[1][2]; afTmp[9]  = aafTSrc[0][2] * aafTSrc[1][0];
  afTmp[10] = aafTSrc[0][0] * aafTSrc[1][1]; afTmp[11] = aafTSrc[0][1] * aafTSrc[1][0];

  // Calculate the second 8 cofactors
  aafDst[2][0] = (afTmp[0]  * aafTSrc[3][1] + afTmp[3]  * aafTSrc[3][2] + afTmp[4]  * aafTSrc[3][3]) - (afTmp[1]  * aafTSrc[3][1] + afTmp[2]  * aafTSrc[3][2] + afTmp[5]  * aafTSrc[3][3]);
  aafDst[2][1] = (afTmp[1]  * aafTSrc[3][0] + afTmp[6]  * aafTSrc[3][2] + afTmp[9]  * aafTSrc[3][3]) - (afTmp[0]  * aafTSrc[3][0] + afTmp[7]  * aafTSrc[3][2] + afTmp[8]  * aafTSrc[3][3]);
  aafDst[2][2] = (afTmp[2]  * aafTSrc[3][0] + afTmp[7]  * aafTSrc[3][1] + afTmp[10] * aafTSrc[3][3]) - (afTmp[3]  * aafTSrc[3][0] + afTmp[6]  * aafTSrc[3][1] + afTmp[11] * aafTSrc[3][3]);
  aafDst[2][3] = (afTmp[5]  * aafTSrc[3][0] + afTmp[8]  * aafTSrc[3][1] + afTmp[11] * aafTSrc[3][2]) - (afTmp[4]  * aafTSrc[3][0] + afTmp[9]  * aafTSrc[3][1] + afTmp[10] * aafTSrc[3][2]);
  aafDst[3][0] = (afTmp[2]  * aafTSrc[2][2] + afTmp[5]  * aafTSrc[2][3] + afTmp[1]  * aafTSrc[2][1]) - (afTmp[4]  * aafTSrc[2][3] + afTmp[0]  * aafTSrc[2][1] + afTmp[3]  * aafTSrc[2][2]);
  aafDst[3][1] = (afTmp[8]  * aafTSrc[2][3] + afTmp[0]  * aafTSrc[2][0] + afTmp[7]  * aafTSrc[2][2]) - (afTmp[6]  * aafTSrc[2][2] + afTmp[9]  * aafTSrc[2][3] + afTmp[1]  * aafTSrc[2][0]);
  aafDst[3][2] = (afTmp[6]  * aafTSrc[2][1] + afTmp[11] * aafTSrc[2][3] + afTmp[3]  * aafTSrc[2][0]) - (afTmp[10] * aafTSrc[2][3] + afTmp[2]  * aafTSrc[2][0] + afTmp[7]  * aafTSrc[2][1]);
  aafDst[3][3] = (afTmp[10] * aafTSrc[2][2] + afTmp[4]  * aafTSrc[2][0] + afTmp[9]  * aafTSrc[2][1]) - (afTmp[8]  * aafTSrc[2][1] + afTmp[11] * aafTSrc[2][2] + afTmp[5]  * aafTSrc[2][0]);

  // Calculate the 1/D value
  fDet = aafTSrc[0][0] * aafDst[0][0] + aafTSrc[0][1] * aafDst[0][1] + aafTSrc[0][2] * aafDst[0][2] + aafTSrc[0][3] * aafDst[0][3];
  fDet = 1 / fDet;

  // Multiply all elements with the 1/D value
  aafMatrix[0][0] = aafDst[0][0] * fDet; aafMatrix[0][1] = aafDst[0][1] * fDet; aafMatrix[0][2] = aafDst[0][2] * fDet; aafMatrix[0][3] = aafDst[0][3] * fDet;
  aafMatrix[1][0] = aafDst[1][0] * fDet; aafMatrix[1][1] = aafDst[1][1] * fDet; aafMatrix[1][2] = aafDst[1][2] * fDet; aafMatrix[1][3] = aafDst[1][3] * fDet;
  aafMatrix[2][0] = aafDst[2][0] * fDet; aafMatrix[2][1] = aafDst[2][1] * fDet; aafMatrix[2][2] = aafDst[2][2] * fDet; aafMatrix[2][3] = aafDst[2][3] * fDet;
  aafMatrix[3][0] = aafDst[3][0] * fDet; aafMatrix[3][1] = aafDst[3][1] * fDet; aafMatrix[3][2] = aafDst[3][2] * fDet; aafMatrix[3][3] = aafDst[3][3] * fDet;
}

void CMatrix3D::Shear(float fX, float fY)
{
  // Temporary variables
  float _sse aafMat[4][4];

  // Create shearing matrix
  aafMat[0][0] = 1; aafMat[0][1] = 0; aafMat[0][2] = fX; aafMat[0][3] = 0;
  aafMat[1][0] = 0; aafMat[1][1] = 1; aafMat[1][2] = fY; aafMat[1][3] = 0;
  aafMat[2][0] = 0; aafMat[2][1] = 0; aafMat[2][2] = 1;  aafMat[2][3] = 0;
  aafMat[3][0] = 0; aafMat[3][1] = 0; aafMat[3][2] = 0;  aafMat[3][3] = 1;

  // Merge with the main matrix
  _MergeMatrix_(aafMat);
}

void CMatrix3D::Scale(float fX, float fY, float fZ)
{
  // Temporary variables
  float _sse aafMat[4][4];

  // Create scalling matrix
  aafMat[0][0] = fX; aafMat[0][1] = 0;  aafMat[0][2] = 0;  aafMat[0][3] = 0;
  aafMat[1][0] = 0;  aafMat[1][1] = fY; aafMat[1][2] = 0;  aafMat[1][3] = 0;
  aafMat[2][0] = 0;  aafMat[2][1] = 0;  aafMat[2][2] = fZ; aafMat[2][3] = 0;
  aafMat[3][0] = 0;  aafMat[3][1] = 0;  aafMat[3][2] = 0;  aafMat[3][3] = 1;

  // Merge with the main matrix
  _MergeMatrix_(aafMat);
}

void CMatrix3D::Rotate(int32_t n32X, int32_t n32Y, int32_t n32Z)
{
  // Temporary variables
  float _sse aafMat[4][4];
  float _sse aafRMatrix[4][4];

  // Create identity matrix
  MakeIdentityMatrix(aafRMatrix);

  // Create Z rotation matrix
  aafMat[0][0] =  COS(n32Z); aafMat[0][1] = SIN(n32Z); aafMat[0][2] = 0; aafMat[0][3] = 0;
  aafMat[1][0] = -SIN(n32Z); aafMat[1][1] = COS(n32Z); aafMat[1][2] = 0; aafMat[1][3] = 0;
  aafMat[2][0] = 0;          aafMat[2][1] = 0;         aafMat[2][2] = 1; aafMat[2][3] = 0;
  aafMat[3][0] = 0;          aafMat[3][1] = 0;         aafMat[3][2] = 0; aafMat[3][3] = 1;
  _MergeMatrices_(aafRMatrix, aafMat);

  // Create X rotation matrix
  aafMat[0][0] = 1; aafMat[0][1] = 0;          aafMat[0][2] = 0;         aafMat[0][3] = 0;
  aafMat[1][0] = 0; aafMat[1][1] =  COS(n32X); aafMat[1][2] = SIN(n32X); aafMat[1][3] = 0;
  aafMat[2][0] = 0; aafMat[2][1] = -SIN(n32X); aafMat[2][2] = COS(n32X); aafMat[2][3] = 0;
  aafMat[3][0] = 0; aafMat[3][1] = 0;          aafMat[3][2] = 0;         aafMat[3][3] = 1;
  _MergeMatrices_(aafRMatrix, aafMat);

  // Create Y rotation matrix
  aafMat[0][0] = COS(n32Y); aafMat[0][1] = 0; aafMat[0][2] = -SIN(n32Y); aafMat[0][3] = 0;
  aafMat[1][0] = 0;         aafMat[1][1] = 1; aafMat[1][2] = 0;          aafMat[1][3] = 0;
  aafMat[2][0] = SIN(n32Y); aafMat[2][1] = 0; aafMat[2][2] = COS(n32Y);  aafMat[2][3] = 0;
  aafMat[3][0] = 0;         aafMat[3][1] = 0; aafMat[3][2] = 0;          aafMat[3][3] = 1;
  _MergeMatrices_(aafRMatrix, aafMat);

  // Merge with the main matrix
  _MergeMatrix_(aafRMatrix);
}

void CMatrix3D::RotateX(int32_t n32X)
{
  // Temporary variables
  float _sse aafMat[4][4];

  // Create and merge X rotation matrix
  aafMat[0][0] = 1; aafMat[0][1] = 0;          aafMat[0][2] = 0;         aafMat[0][3] = 0;
  aafMat[1][0] = 0; aafMat[1][1] =  COS(n32X); aafMat[1][2] = SIN(n32X); aafMat[1][3] = 0;
  aafMat[2][0] = 0; aafMat[2][1] = -SIN(n32X); aafMat[2][2] = COS(n32X); aafMat[2][3] = 0;
  aafMat[3][0] = 0; aafMat[3][1] = 0;          aafMat[3][2] = 0;         aafMat[3][3] = 1;
  _MergeMatrix_(aafMat);
}

void CMatrix3D::RotateY(int32_t n32Y)
{
  // Temporary variables
  float _sse aafMat[4][4];

  // Create and merge Y rotation matrix
  aafMat[0][0] = COS(n32Y); aafMat[0][1] = 0; aafMat[0][2] = -SIN(n32Y); aafMat[0][3] = 0;
  aafMat[1][0] = 0;         aafMat[1][1] = 1; aafMat[1][2] = 0;          aafMat[1][3] = 0;
  aafMat[2][0] = SIN(n32Y); aafMat[2][1] = 0; aafMat[2][2] = COS(n32Y);  aafMat[2][3] = 0;
  aafMat[3][0] = 0;         aafMat[3][1] = 0; aafMat[3][2] = 0;          aafMat[3][3] = 1;
  _MergeMatrix_(aafMat);
}

void CMatrix3D::RotateZ(int32_t n32Z)
{
  // Temporary variables
  float _sse aafMat[4][4];

  // Create and merge Z rotation matrix
  aafMat[0][0] =  COS(n32Z); aafMat[0][1] = SIN(n32Z); aafMat[0][2] = 0; aafMat[0][3] = 0;
  aafMat[1][0] = -SIN(n32Z); aafMat[1][1] = COS(n32Z); aafMat[1][2] = 0; aafMat[1][3] = 0;
  aafMat[2][0] = 0;          aafMat[2][1] = 0;         aafMat[2][2] = 1; aafMat[2][3] = 0;
  aafMat[3][0] = 0;          aafMat[3][1] = 0;         aafMat[3][2] = 0; aafMat[3][3] = 1;
  _MergeMatrix_(aafMat);
}

void CMatrix3D::Translate(float fX, float fY, float fZ)
{
  // Temporary variables
  float _sse aafMat[4][4];

  // Create translation matrix
  aafMat[0][0] = 1;  aafMat[0][1] = 0;  aafMat[0][2] = 0;  aafMat[0][3] = 0;
  aafMat[1][0] = 0;  aafMat[1][1] = 1;  aafMat[1][2] = 0;  aafMat[1][3] = 0;
  aafMat[2][0] = 0;  aafMat[2][1] = 0;  aafMat[2][2] = 1;  aafMat[2][3] = 0;
  aafMat[3][0] = fX; aafMat[3][1] = fY; aafMat[3][2] = fZ; aafMat[3][3] = 1;

  // Merge with the main matrix
  _MergeMatrix_(aafMat);
}

} // namespace vge
