// ############################################################################################
//  Panel3D.cc
//  Implementation of CPanel3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/LookUpTables.h"
#include "FixedPointConsts.h"
#include "FoggedCeilingLineSegment.h"
#include "Panel3D.h"
#include "UnfoggedCeilingLineSegment.h"


namespace vge {

//
// External functions prototpes
//
void _UFClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep);
void _UGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                        int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                        int32_t &n32B, int32_t &n32BStep);
void _TFClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                        int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep);
void _TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                        int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep,
                        int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                        int32_t &n32B, int32_t &n32BStep);

void _FogUF_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                               int32_t &n32F, int32_t &n32FStep);
void _FogUG_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                              int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                              int32_t &n32B, int32_t &n32BStep, int32_t &n32F, int32_t &n32FStep);
void _FogTF_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                              int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep,
                              int32_t &n32F, int32_t &n32FStep);
void _FogTG_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                              int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep,
                              int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                              int32_t &n32B, int32_t &n32BStep, int32_t &n32F, int32_t &n32FStep);


//
// The 3D panel class
//
void CPanel3D::SetTextureCoordinates(TTexUV atUV[4])
{
  memcpy(&_tP.atUV, atUV, sizeof(_tP.atUV[0]) * 4);
}

void CPanel3D::SetTextureCoordinates(float fU1, float fV1, float fU2, float fV2, float fU3, float fV3, float fU4, float fV4)
{
  _tP.atUV[0].fU = fU1; _tP.atUV[0].fV = fV1;
  _tP.atUV[1].fU = fU2; _tP.atUV[1].fV = fV2;
  _tP.atUV[2].fU = fU3; _tP.atUV[2].fV = fV3;
  _tP.atUV[3].fU = fU4; _tP.atUV[3].fV = fV4;
}

bool CPanel3D::IsVisible2D()
{
  // Temporary variables
  uint32_t un32XMinInVis = 0, un32YMinInVis = 0, un32XMaxInVis = 0, un32YMaxInVis = 0;
  uint8_t  un8Cnt;

  // Make sure the panel has more than two points
  if(_tP.un8SPCount < 3) {
    _tP.un32Invisible = 4;
    return(false);
  }

  // Determine the location of panel's 2D points
  for(un8Cnt = 0; un8Cnt < _tP.un8SPCount; un8Cnt++) {
    if(_tP.atSPoint[un8Cnt].n32X < C3DEnv::Active3DEnv()->_n32MinX)
      un32XMinInVis++;
    else if(_tP.atSPoint[un8Cnt].n32X > C3DEnv::Active3DEnv()->_n32MaxX)
      un32XMaxInVis++;

    if(_tP.atSPoint[un8Cnt].n32Y < C3DEnv::Active3DEnv()->_n32MinY)
      un32YMinInVis++;
    else if(_tP.atSPoint[un8Cnt].n32Y > C3DEnv::Active3DEnv()->_n32MaxY)
      un32YMaxInVis++;
  }

  // Checking
  if((un32XMinInVis >= _tP.un8SPCount) ||
     (un32YMinInVis >= _tP.un8SPCount) ||
     (un32XMaxInVis >= _tP.un8SPCount) ||
     (un32YMaxInVis >= _tP.un8SPCount)) {
    _tP.un32Invisible = max(max(un32XMinInVis, un32YMinInVis),
                            max(un32XMaxInVis, un32YMaxInVis));
    return(false);
  }
  return(true);
}

bool CPanel3D::_CalcBFace_()
{
  // For covenience
  const float &wx = _tP.apcsVPoint[0]->_tWorld.fX;
  const float &wy = _tP.apcsVPoint[0]->_tWorld.fY;
  const float &wz = _tP.apcsVPoint[0]->_tWorld.fZ;
  const float &nx = _tP.tBNormal.fXt;
  const float &ny = _tP.tBNormal.fYt;
  const float &nz = _tP.tBNormal.fZt;

  // Generate points
  const CPoint3D wp(wx, wy, wz);
  const CPoint3D np(nx-wx, ny-wy, nz-wz);

  // Calculate the cosine value of the angle between the view vector
  // and the transformed panel's normal (without normalization)
  float angle = wp.DotUnit(np);

  // If back-faced, perform more accurate test
  if(angle > 0.0f) {
    angle /= (wp.Mag() * np.Mag()); //VPrint("%+05.2f\n", angle);
    if(angle > 0.1f) {
      _tP.un32Invisible = 1; //uint32_t(angle * 5);
      return(false);
    }
  }
  return(true);
}

bool CPanel3D::_CheckExtents_()
{
  // Temporary variables
  bool    bVis = false;
  uint8_t un8Cnt;
  int32_t n32SumZ;
  float   fMinZ;

  // Check againts MinZ
  for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
  if(_tP.apcsVPoint[un8Cnt]->_tWorld.fZ > C3DEnv::Active3DEnv()->_un32MinZ) {
    bVis = true;
    _tP.un32Invisible = 0;
    break;
  }

  // Still visible, check againts MaxZ
  if(bVis) {
    fMinZ = _tP.apcsVPoint[0]->_tWorld.fZ;
    for(un8Cnt = 0; un8Cnt < 4; un8Cnt++)
      if(_tP.apcsVPoint[un8Cnt]->_tWorld.fZ < fMinZ) fMinZ = _tP.apcsVPoint[un8Cnt]->_tWorld.fZ;

    if(fMinZ > C3DEnv::Active3DEnv()->_un32MaxZ) {
      bVis = false;
      _tP.un32Invisible = uint32_t(fMinZ - C3DEnv::Active3DEnv()->_un32MaxZ) >> 6;
    }
  }
  // Not visible
  else {
    n32SumZ = int32_t(_tP.apcsVPoint[0]->_tWorld.fZ + _tP.apcsVPoint[1]->_tWorld.fZ +
                      _tP.apcsVPoint[2]->_tWorld.fZ + _tP.apcsVPoint[3]->_tWorld.fZ);

    _tP.un32Invisible = uint32_t((n32SumZ < 0) ? -n32SumZ : n32SumZ) >> 9;
  }

  // Return the value
  return(bVis);
}

void CPanel3D::_CalcNormal_()
{
  // Create a list of unique vertices
  int      range = 0;
  CPoint3D unique[4];
  for(int i = 0; i < 4; ++i) {
    const CPoint3D &tmp = *_tP.apcsVPoint[i];
    if(!range)
      unique[range++] = tmp;
    else if(UniqueVert(tmp, unique, range))
      unique[range++] = tmp;
  }

  // For convenience in the next equations
  const float &x1 = unique[0]._tLocal.fX;
  const float &y1 = unique[0]._tLocal.fY;
  const float &z1 = unique[0]._tLocal.fZ;
  const float &x2 = unique[1]._tLocal.fX;
  const float &y2 = unique[1]._tLocal.fY;
  const float &z2 = unique[1]._tLocal.fZ;
  const float &x3 = unique[2]._tLocal.fX;
  const float &y3 = unique[2]._tLocal.fY;
  const float &z3 = unique[2]._tLocal.fZ;

  // Use cross product to determine plane's orientation
#if 1
  const float a = (y2-y1)*(z3-z1) - (z2-z1)*(y3-y1);
  const float b = (z2-z1)*(x3-x1) - (x2-x1)*(z3-z1);
  const float c = (x2-x1)*(y3-y1) - (y2-y1)*(x3-x1);
#else
  const float a = y1*(z2-z3) + y2*(z3-z1) + y3*(z1-z2);
  const float b = z1*(x2-x3) + z2*(x3-x1) + z3*(x1-x2);
  const float c = x1*(y2-y3) + x2*(y3-y1) + x3*(y1-y2);
#endif

  // Caclculate the distance to the vector
  const float dist = sqrt(a*a + b*b + c*c);

  // Create the normalized normal
  _tP.tNormal.fX  = a / dist;
  _tP.tNormal.fY  = b / dist;
  _tP.tNormal.fZ  = c / dist;
  _tP.tBNormal.fX = a + _tP.apcsVPoint[0]->_tLocal.fX;
  _tP.tBNormal.fY = b + _tP.apcsVPoint[0]->_tLocal.fY;
  _tP.tBNormal.fZ = c + _tP.apcsVPoint[0]->_tLocal.fZ;
  //VPrint("N: %+05.2f, %+05.2f, %+05.2f\n", _tP.tNormal.fX, _tP.tNormal.fY, _tP.tNormal.fZ);
  //VPrint("B: %+05.2f, %+05.2f, %+05.2f\n", _tP.tBNormal.fX, _tP.tBNormal.fY, _tP.tBNormal.fZ);
}

void CPanel3D::_CalcRadius_()
{
  // Temporary variables
  CPoint3D acsTmp[4], csCenter;
  float          afDist[4];
  uint8_t        un8Cnt;

  // Create a temporary vertex list & calculate the center of panel
  for(un8Cnt = 0; un8Cnt < 4; un8Cnt++) {
    acsTmp[un8Cnt] = *_tP.apcsVPoint[un8Cnt];
    csCenter      += acsTmp[un8Cnt];
  }
  csCenter /= 4.0f;
  _tP.afCenter[0] = csCenter._tLocal.fX;
  _tP.afCenter[1] = csCenter._tLocal.fY;
  _tP.afCenter[2] = csCenter._tLocal.fZ;

  // Calculate the distance to each of the vertices
  for(un8Cnt = 0; un8Cnt < 4; un8Cnt++) {
    acsTmp[un8Cnt] -= csCenter;
    afDist[un8Cnt]  = acsTmp[un8Cnt].Mag();
  }

  // Determine the maximum distance as the radius
  _tP.fRadius = afDist[0];
  for(un8Cnt = 1; un8Cnt < 4; un8Cnt++)
    if(afDist[un8Cnt] > _tP.fRadius) _tP.fRadius = afDist[un8Cnt];
}

void CPanel3D::_FSCalcIntenOmni_(const TOmniLight *ptVOmniLight)
{
  // Temporary variables
  float fCos;
  float afXYZ[4];

  // Calculate the center of panel
  afXYZ[0] = _tP.apcsVPoint[0]->_tLocal.fX;
  afXYZ[1] = _tP.apcsVPoint[0]->_tLocal.fY;
  afXYZ[2] = _tP.apcsVPoint[0]->_tLocal.fZ;
  afXYZ[0] += _tP.apcsVPoint[1]->_tLocal.fX;
  afXYZ[1] += _tP.apcsVPoint[1]->_tLocal.fY;
  afXYZ[2] += _tP.apcsVPoint[1]->_tLocal.fZ;
  afXYZ[0] += _tP.apcsVPoint[2]->_tLocal.fX;
  afXYZ[1] += _tP.apcsVPoint[2]->_tLocal.fY;
  afXYZ[2] += _tP.apcsVPoint[2]->_tLocal.fZ;
  afXYZ[0] += _tP.apcsVPoint[3]->_tLocal.fX;
  afXYZ[1] += _tP.apcsVPoint[3]->_tLocal.fY;
  afXYZ[2] += _tP.apcsVPoint[3]->_tLocal.fZ;
  afXYZ[0] /= 4; afXYZ[1] /= 4; afXYZ[2] /= 4;

  // Use the supplied omni light
  if(ptVOmniLight) {
    fCos = VCalcLightCosValue(&_tP.tNormal.fX, afXYZ, &ptVOmniLight->fX);
    _tP.tOmni.un8R = uint8_t(ptVOmniLight->un8R * fCos);
    _tP.tOmni.un8G = uint8_t(ptVOmniLight->un8G * fCos);
    _tP.tOmni.un8B = uint8_t(ptVOmniLight->un8B * fCos);
  }
  // Use the default omni light from the active 3D environment
  else {
    fCos = VCalcLightCosValue(&_tP.tNormal.fX, afXYZ, &C3DEnv::Active3DEnv()->_tDefOLight.fX);
    _tP.tOmni.un8R = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8R * fCos);
    _tP.tOmni.un8G = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8G * fCos);
    _tP.tOmni.un8B = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8B * fCos);
  }
}

void CPanel3D::_GSCalcIntenOmni_(const TOmniLight *ptVOmniLight, bool bUseAvgNormal)
{
  // Temporary variables
  float fCos;
  uint8_t un8Cnt;

  // Use the supplied omni light
  if(ptVOmniLight) {
    for(un8Cnt = 0; un8Cnt < 4; un8Cnt++) {
      if(bUseAvgNormal)
        fCos = VCalcLightCosValue(&_tP.apcsVPoint[un8Cnt]->_tAvgN.fX, &_tP.apcsVPoint[un8Cnt]->_tLocal.fX, &ptVOmniLight->fX);
      else
        fCos = VCalcLightCosValue(&_tP.tNormal.fX, &_tP.apcsVPoint[un8Cnt]->_tLocal.fX, &ptVOmniLight->fX);
      _tP.atOmni[un8Cnt].un8R = uint8_t(ptVOmniLight->un8R * fCos);
      _tP.atOmni[un8Cnt].un8G = uint8_t(ptVOmniLight->un8G * fCos);
      _tP.atOmni[un8Cnt].un8B = uint8_t(ptVOmniLight->un8B * fCos);
    }
  }
  // Use the default omni light from the active 3D environment
  else {
    for(un8Cnt = 0; un8Cnt < 4; un8Cnt++) {
      if(bUseAvgNormal)
        fCos = VCalcLightCosValue(&_tP.apcsVPoint[un8Cnt]->_tAvgN.fX, &_tP.apcsVPoint[un8Cnt]->_tLocal.fX, &C3DEnv::Active3DEnv()->_tDefOLight.fX);
      else
        fCos = VCalcLightCosValue(&_tP.tNormal.fX, &_tP.apcsVPoint[un8Cnt]->_tLocal.fX, &C3DEnv::Active3DEnv()->_tDefOLight.fX);
      _tP.atOmni[un8Cnt].un8R = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8R * fCos);
      _tP.atOmni[un8Cnt].un8G = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8G * fCos);
      _tP.atOmni[un8Cnt].un8B = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8B * fCos);
    }
  }
}

void CPanel3D::_DynFSCalcIntenOmni_(const TOmniLight *ptVOmniLight)
{
  // Temporary variables
  float fCos;
  float afXYZ[4];

  // Calculate the center of panel
  afXYZ[0] = _tP.apcsVPoint[0]->_tWorld.fX;
  afXYZ[1] = _tP.apcsVPoint[0]->_tWorld.fY;
  afXYZ[2] = _tP.apcsVPoint[0]->_tWorld.fZ;
  afXYZ[0] += _tP.apcsVPoint[1]->_tWorld.fX;
  afXYZ[1] += _tP.apcsVPoint[1]->_tWorld.fY;
  afXYZ[2] += _tP.apcsVPoint[1]->_tWorld.fZ;
  afXYZ[0] += _tP.apcsVPoint[2]->_tWorld.fX;
  afXYZ[1] += _tP.apcsVPoint[2]->_tWorld.fY;
  afXYZ[2] += _tP.apcsVPoint[2]->_tWorld.fZ;
  afXYZ[0] += _tP.apcsVPoint[3]->_tWorld.fX;
  afXYZ[1] += _tP.apcsVPoint[3]->_tWorld.fY;
  afXYZ[2] += _tP.apcsVPoint[3]->_tWorld.fZ;
  afXYZ[0] /= 4; afXYZ[1] /= 4; afXYZ[2] /= 4;

  // Use the supplied omni light
  if(ptVOmniLight) {
    fCos = VCalcLightCosValue(&_tP.tNormal.fX, afXYZ, &ptVOmniLight->fX);
    _tP.tOmni.un8R = uint8_t(ptVOmniLight->un8R * fCos);
    _tP.tOmni.un8G = uint8_t(ptVOmniLight->un8G * fCos);
    _tP.tOmni.un8B = uint8_t(ptVOmniLight->un8B * fCos);
  }
  // Use the default omni light from the active 3D environment
  else {
    fCos = VCalcLightCosValue(&_tP.tNormal.fX, afXYZ, &C3DEnv::Active3DEnv()->_tDefOLight.fX);
    _tP.tOmni.un8R = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8R * fCos);
    _tP.tOmni.un8G = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8G * fCos);
    _tP.tOmni.un8B = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8B * fCos);
  }
}

void CPanel3D::_DynGSCalcIntenOmni_(const TOmniLight *ptVOmniLight)
{
  // Temporary variables
  float fCos;
  uint8_t un8Cnt;

  // Use the supplied omni light
  if(ptVOmniLight) {
    for(un8Cnt = 0; un8Cnt < 4; un8Cnt++) {
      fCos = VCalcLightCosValue(&_tP.tNormal.fX, &_tP.apcsVPoint[un8Cnt]->_tWorld.fX, &ptVOmniLight->fX);
      _tP.atOmni[un8Cnt].un8R = uint8_t(ptVOmniLight->un8R * fCos);
      _tP.atOmni[un8Cnt].un8G = uint8_t(ptVOmniLight->un8G * fCos);
      _tP.atOmni[un8Cnt].un8B = uint8_t(ptVOmniLight->un8B * fCos);
    }
  }
  // Use the default omni light from the active 3D environment
  else {
    for(un8Cnt = 0; un8Cnt < 4; un8Cnt++) {
      fCos = VCalcLightCosValue(&_tP.tNormal.fX, &_tP.apcsVPoint[un8Cnt]->_tWorld.fX, &C3DEnv::Active3DEnv()->_tDefOLight.fX);
      _tP.atOmni[un8Cnt].un8R = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8R * fCos);
      _tP.atOmni[un8Cnt].un8G = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8G * fCos);
      _tP.atOmni[un8Cnt].un8B = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8B * fCos);
    }
  }
}

void CPanel3D::_GSCalcInten_()
{
  for(uint8_t un8Cnt = 0; un8Cnt < 4; un8Cnt++) {
    _tP.atTotal[un8Cnt].un8R = (_tP.tAmbient.un8R + _tP.atOmni[un8Cnt].un8R) >> 1;
    _tP.atTotal[un8Cnt].un8G = (_tP.tAmbient.un8G + _tP.atOmni[un8Cnt].un8G) >> 1;
    _tP.atTotal[un8Cnt].un8B = (_tP.tAmbient.un8B + _tP.atOmni[un8Cnt].un8B) >> 1;
  }
}

} // namespace vge
