// ############################################################################################
//  AALSAEnvironment.cc
//  Implementation of CAlsaAudioEnv class using ALSA audio system
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#include "../../VConfig.h"
#ifdef __USE_ALSA_AUDIO__

// VortexGE headers
#include "../MidiPlayer.h"
#include "AALSASoundBuffer.h"

// ALSA headers
#include <asoundlib.h>


namespace vge {

//
// The ALSA audio environment implementation
//
VCLASS_SINGLETON_IMPLEMENT(CAlsaAudioEnv)

CAlsaAudioEnv *CAlsaAudioEnv::_instance = NULL;

CAlsaAudioEnv::CAlsaAudioEnv()
{
  if(!CAlsaAudioEnv::_instance) {
    CAudioEnvCImp::_instAudioEnv = this;
    CAlsaAudioEnv::_instance = this;
  }

  CAudioEnvCImp::_tCD = &_tD;

  memset(_atPSB, 0, sizeof(_atPSB[0]) * _MAX_SB_);
  memset(&_tD, 0, sizeof(_tD));
}

CAlsaAudioEnv::~CAlsaAudioEnv() throw()
{
  // Terminate threads
  try {
    _csTWriter.Stop();
  }
  catch(...) {
    _csTWriter.Kill();
  }

  // Close audio device
  if(_tD.deviceHandle) {
    snd_pcm_drop(_tD.deviceHandle);
    snd_pcm_close(_tD.deviceHandle);
    _tD.deviceHandle = NULL;
  }
  if(_tD.hwParams) {
    snd_pcm_hw_params_free(_tD.hwParams);
    _tD.hwParams = NULL;
  }

  // Free memory
  if(_tD.pucPBuff) free(_tD.pucPBuff);

  // Set reference to NULL
  if(CAlsaAudioEnv::_instance == this) CAlsaAudioEnv::_instance = NULL;
}

void CAlsaAudioEnv::DoQuickInit(const char *pcDeviceName)
{
  if(pcDeviceName)
    OpenDevice(pcDeviceName);
  else {
    char acBuff[PATH_MAX];
    QueryDeviceName(acBuff, sizeof(acBuff));
    OpenDevice(acBuff);
  }
  InitializeDevice();
  ActivateDevice();
}

// NOTE: 1. If pcResult is NULL, this function will only check the
//          availability of the device.
//       2. In case of the system has multiple devices, only the 1st
//          device name will be returned. If the first device can't
//          be opened, OpenDevice() will adjust the device name to
//          open the 2nd, 3rd, and so.
size_t CAlsaAudioEnv::QueryDeviceName(char *pcResult, size_t bufferSize)
{
  // Empty the destination string
  if(pcResult) pcResult[0] = 0;

  // Search for /dev/snd/pcmC*D*p
  for(size_t c = 0; c <= 9; c++) {
    for(size_t d = 0; d <= 9; d++) {

      char devName[PATH_MAX];
      snprintf(devName, sizeof(devName), "/dev/snd/pcmC%zdD%zdp", c, d);

      struct stat tStat;
      if((stat(devName, &tStat) == 0) && S_ISCHR(tStat.st_mode)) {
        char alsaName[PATH_MAX];
        snprintf(alsaName, sizeof(alsaName), "hw:%zd,%zd", c, d);

        if(pcResult && bufferSize) safe_strcpy(pcResult, alsaName, bufferSize);
        return(strlen(alsaName) + 1);
      }
    }
  }

  // Not found
  VError(VERR_NotFound, _NFL_, "Couldn't not found any ALSA audio device in this system");
}

bool CAlsaAudioEnv::_AdjustDeviceName_(char *pcDeviceName)
{
  char *c = NULL;
  char *d = NULL;
  if(strncmp(pcDeviceName, "hw:", 3) == 0) {
    // The name is in the from of hw:*,* 0123456
    c = pcDeviceName + 3;
    d = pcDeviceName + 5;
  }
  else if(strncmp(pcDeviceName, "plughw:", 7) == 0) {
    // The name is in the from of plughw:*,*
    c = pcDeviceName + 7;
    d = pcDeviceName + 9;
  }
  else
    return(false);

  // Check
  if(*c=='9' && *d=='9') return(false);

  // Adjust the name
  switch(*d) {
    case '0': *d = '1'; return(true);
    case '1': *d = '2'; return(true);
    case '2': *d = '3'; return(true);
    case '3': *d = '4'; return(true);
    case '4': *d = '5'; return(true);
    case '5': *d = '6'; return(true);
    case '6': *d = '7'; return(true);
    case '7': *d = '8'; return(true);
    case '8': *d = '9'; return(true);
    default : *d = '0';
  };
  switch(*c) {
    case '0': *c = '1'; break;
    case '1': *c = '2'; break;
    case '2': *c = '3'; break;
    case '3': *c = '4'; break;
    case '4': *c = '5'; break;
    case '5': *c = '6'; break;
    case '6': *c = '7'; break;
    case '7': *c = '8'; break;
    case '8': *c = '9'; break;
    default : return(false);
  };
  return(true);
}

void CAlsaAudioEnv::OpenDevice(const char *pcDeviceName)
{
  // Check
  if(_tD.deviceHandle) return;

  // If the user have a preferred device name set in the environment variable,
  // try that first
  char *pcEnv = getenv("VGE_AUDIO_DEVICE");
  if(pcEnv) {
    // Try to open in nonblocking mode
    int32_t err = snd_pcm_open(&_tD.deviceHandle, pcEnv, SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK | SND_PCM_ASYNC);
    if(err >= 0) {
      // Reopen in blocking mode
      snd_pcm_close(_tD.deviceHandle);
      err = snd_pcm_open(&_tD.deviceHandle, pcEnv, SND_PCM_STREAM_PLAYBACK, SND_PCM_ASYNC);
      if(err < 0) {
        string msg = "Couldn't reopen ALSA device ";
        msg += pcEnv;
        msg += "("; msg += snd_strerror(err); msg += ")";
        VError(VERR_OpenDevice, _NFL_, msg.c_str());
      }
      // Copy the device name and finish
      safe_strcpy(_tD.acDevName, pcEnv, sizeof(_tD.acDevName));
      return;
    }
  }

  // Try to open the device given in the pcDeviceName parameter
  // in nonblocking mode
  char acDevName[PATH_MAX];
  safe_strcpy(acDevName, pcDeviceName, sizeof(acDevName));
  int32_t ret = snd_pcm_open(&_tD.deviceHandle, acDevName,
                             SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK | SND_PCM_ASYNC);

  // If can't be opened, try to open the next device
  if(ret < 0) {
    // Adjust the device name and try to open it
    while(_AdjustDeviceName_(acDevName)) {
      ret = snd_pcm_open(&_tD.deviceHandle, acDevName,
                         SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK | SND_PCM_ASYNC);
      if(ret >= 0) break;
    }
    // No device can be opened ?
    if(ret < 0)
      VError(VERR_OpenDevice, _NFL_, "No usable ALSA audio device could be found");
  }

  // Close the device and reopen in blocking mode
  snd_pcm_close(_tD.deviceHandle);
  ret = snd_pcm_open(&_tD.deviceHandle, acDevName,
                     SND_PCM_STREAM_PLAYBACK, SND_PCM_ASYNC);
  if(ret < 0) {
    string msg = "Couldn't reopen ALSA device ";
    msg += acDevName;
    msg += "("; msg += snd_strerror(ret); msg += ")";
    VError(VERR_OpenDevice, _NFL_, msg.c_str());
  }

  // Copy the device name
  safe_strcpy(_tD.acDevName, acDevName, sizeof(_tD.acDevName));
}

static snd_pcm_format_t _VSBF2ALSA_(ESoundBufferFormat eFormat)
{
  switch(eFormat) {
    case ESB_Undefined     : return(SND_PCM_FORMAT_S16_LE); // If not defined, we prefer 16-bit signed LE
    case ESB_Unknown       : return(SND_PCM_FORMAT_S16_LE); // If unknown, we prefer 16-bit signed LE
    case ESB_Signed8       : return(SND_PCM_FORMAT_S8);
    case ESB_Unsigned8     : return(SND_PCM_FORMAT_U8);
    case ESB_Signed16LE    : return(SND_PCM_FORMAT_S16_LE);
    case ESB_Unsigned16LE  : return(SND_PCM_FORMAT_U16_LE);
    case ESB_Signed16BE    : return(SND_PCM_FORMAT_S16_BE);
    case ESB_Unsigned16BE  : return(SND_PCM_FORMAT_U16_BE);
    default             : VError(VERR_InvalidParameter, _NFL_, "Invalid ESoundBufferFormat");
  }
}

static ESoundBufferFormat _ALSA2VSBF_(snd_pcm_format_t eFormat)
{
  switch(eFormat) {
    case SND_PCM_FORMAT_S8     : return(ESB_Signed8);
    case SND_PCM_FORMAT_U8     : return(ESB_Unsigned8);
    case SND_PCM_FORMAT_S16_LE : return(ESB_Signed16LE);
    case SND_PCM_FORMAT_U16_LE : return(ESB_Unsigned16LE);
    case SND_PCM_FORMAT_S16_BE : return(ESB_Signed16BE);
    case SND_PCM_FORMAT_U16_BE : return(ESB_Unsigned16BE);
    default                    : return(ESB_Unknown);
  }
}

void CAlsaAudioEnv::InitializeDevice(const TAudioSettings *ptPreferred, TAudioSettings *ptRealized)
{
  // Temporary variables
  TAudioSettings tR, tP = { ESB_Undefined, 44100, 2 };
  TAudioSettings *ptR, *ptP;

  // Check
  if(!_tD.deviceHandle)
    VError(VERR_DeviceNotOpenend, _NFL_, "Audio device is not opened");

  // Preparations
  if(ptPreferred)
    ptP = const_cast<TAudioSettings*>(ptPreferred);
  else
    ptP = &tP;
  if(ptRealized)
    ptR = ptRealized;
  else
    ptR = &tR;

  // ..... Try to set the requested format
  int32_t err;

  if((err = snd_pcm_hw_params_malloc(&_tD.hwParams)) < 0) {
    string msg = "Couldn't allocate ALSA HW parameter structure ";
    msg += "("; msg += snd_strerror(err); msg += ")";
    VError(VERR_NotEnoughMemory, _NFL_, msg.c_str());
  }

  if((err = snd_pcm_hw_params_any(_tD.deviceHandle, _tD.hwParams)) < 0) {
    string msg = "Couldn't intilialize ALSA HW parameter structure ";
    msg += "("; msg += snd_strerror(err); msg += ")";
    VError(VERR_NotInitialized, _NFL_, msg.c_str());
  }

  if((err = snd_pcm_hw_params_set_access(_tD.deviceHandle, _tD.hwParams, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
    string msg = "Couldn't set ALSA HW access mode ";
    msg += "("; msg += snd_strerror(err); msg += ")";
    VError(VERR_NotSupported, _NFL_, msg.c_str());
  }

  snd_pcm_format_t fmt = _VSBF2ALSA_(ptP->eFormat);
  if((err = snd_pcm_hw_params_set_format(_tD.deviceHandle, _tD.hwParams, fmt)) < 0) {
    string msg = "Couldn't set ALSA HW sample format ";
    msg += "("; msg += snd_strerror(err); msg += ")";
    VError(VERR_NotSupported, _NFL_, msg.c_str());
  }
  ptR->eFormat = _ALSA2VSBF_(fmt);

  uint32_t frq = ptP->n32Frequency;
  if((err = snd_pcm_hw_params_set_rate_near(_tD.deviceHandle, _tD.hwParams, &frq, 0)) < 0) {
    string msg = "Couldn't set ALSA HW sample rate ";
    msg += "("; msg += snd_strerror(err); msg += ")";
    VError(VERR_NotSupported, _NFL_, msg.c_str());
  }
  ptR->n32Frequency = frq;

  uint32_t chn = ptP->n32Channels;
  if((err = snd_pcm_hw_params_set_channels(_tD.deviceHandle, _tD.hwParams, chn)) < 0) {
    string msg = "Couldn't set ALSA HW channels ";
    msg += "("; msg += snd_strerror(err); msg += ")";
    VError(VERR_NotSupported, _NFL_, msg.c_str());
  }
  if((err = snd_pcm_hw_params_get_channels(_tD.hwParams, &chn)) < 0) {
    string msg = "Couldn't get ALSA HW channels ";
    msg += "("; msg += snd_strerror(err); msg += ")";
    VError(VERR_NotSupported, _NFL_, msg.c_str());
  }
  ptR->n32Channels = chn;

  uint32_t btime = 200000;
  if((err = snd_pcm_hw_params_set_buffer_time_near(_tD.deviceHandle, _tD.hwParams, &btime, 0)) < 0) {
    string msg = "Couldn't set ALSA HW buffer time ";
    msg += "("; msg += snd_strerror(err); msg += ")";
    VError(VERR_NotSupported, _NFL_, msg.c_str());
  }

  uint32_t ptime = btime / 4;
  if((err = snd_pcm_hw_params_set_period_time_near(_tD.deviceHandle, _tD.hwParams, &ptime, 0)) < 0) {
    string msg = "Couldn't set ALSA HW period time ";
    msg += "("; msg += snd_strerror(err); msg += ")";
    VError(VERR_NotSupported, _NFL_, msg.c_str());
  }

  if((err = snd_pcm_hw_params(_tD.deviceHandle, _tD.hwParams)) < 0) {
    string msg = "Couldn't realize ALSA HW paramaters ";
    msg += "("; msg += snd_strerror(err); msg += ")";
    VError(VERR_NotSupported, _NFL_, msg.c_str());
  }

  // Prepare for playback
  if((err = snd_pcm_prepare(_tD.deviceHandle)) < 0) {
    string msg = "Couldn't prepare ALSA HW interface for use ";
    msg += "("; msg += snd_strerror(err); msg += ")";
    VError(VERR_NotSupported, _NFL_, msg.c_str());
  }

  // Check and copy formats
  if(ptR->eFormat == ESB_Unsigned8) {
    _tD.tFormat.un8Bits   = 8;
    _tD.tFormat.un8Sign   = 0;
    _tD.tFormat.un8Endian = 0;
  }
  else if(ptR->eFormat == ESB_Signed8) {
    _tD.tFormat.un8Bits   = 8;
    _tD.tFormat.un8Sign   = 1;
    _tD.tFormat.un8Endian = 0;
  }
  else if(ptR->eFormat == ESB_Unsigned16LE) {
    _tD.tFormat.un8Bits   = 16;
    _tD.tFormat.un8Sign   = 0;
    _tD.tFormat.un8Endian = 0;
  }
  else if(ptR->eFormat == ESB_Unsigned16BE) {
    _tD.tFormat.un8Bits   = 16;
    _tD.tFormat.un8Sign   = 0;
    _tD.tFormat.un8Endian = 1;
  }
  else if(ptR->eFormat == ESB_Signed16LE) {
    _tD.tFormat.un8Bits   = 16;
    _tD.tFormat.un8Sign   = 1;
    _tD.tFormat.un8Endian = 0;
  }
  else if(ptR->eFormat == ESB_Signed16BE) {
    _tD.tFormat.un8Bits   = 16;
    _tD.tFormat.un8Sign   = 1;
    _tD.tFormat.un8Endian = 1;
  }
  else
    VError(VERR_NotSupported, _NFL_,
           "The audio device does not support any audio format compatible with this application");

  _tD.tFormat.un8Channels   = ptR->n32Channels;
  _tD.tFormat.un32Frequency = ptR->n32Frequency;

  uint32_t fs1 = uint32_t(float(_tD.tFormat.un8Channels * (_tD.tFormat.un8Bits / 8) * _tD.tFormat.un32Frequency) * (float(ptime) / 1000000.0f));
  uint32_t fs2 = 2048 * _tD.tFormat.un8Channels * (_tD.tFormat.un8Bits / 8);
  _tD.un32FragSize  = min(fs1, fs2);

  _tD.un32FragSize2 = _tD.un32FragSize >> 1;
  _tD.un32FragSize4 = _tD.un32FragSize >> 2;
  _tD.un32PBuffSize = _tD.un32FragSize * 4;
  _tD.un32WriteIncF = _tD.tFormat.un8Channels * (_tD.tFormat.un8Bits / 8);
  _tD.un32WriteSize = _tD.un32FragSize / _tD.un32WriteIncF;
  _eFormat          = ptR->eFormat;

  // Allocate memory for primary buffer
  _tD.pucPBuff = reinterpret_cast<u_char*>(malloc(_tD.un32PBuffSize));
  if(_tD.pucPBuff == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for primary sound buffer");
  _tD.pucPBuffRead  = _tD.pucPBuff;
  _tD.pucPBuffWrite = _tD.pucPBuff;
  _tD.pucPBuffMax   = _tD.pucPBuff + _tD.un32PBuffSize - _tD.un32FragSize;
  _tD.un8ToWrite    = 0;
  memset(_tD.pucPBuff, 0, _tD.un32PBuffSize);

  // Initialize threads
  _csTWriter.Initialize(CAlsaAudioEnv::_ThreadProc, this);

  // Initialize sound mixing function
  // Assuming that this system is a little-endian (x86)
  //! \TODO Convert it to table !!!
  if(_tD.tFormat.un8Channels == 1) {
    if(_tD.tFormat.un8Bits == 8) {
      if(_tD.tFormat.un8Sign == 0)     _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerU8m;   // U8 mono
      else                             _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerS8m;   // S8 mono
    }
    else {
      if(_tD.tFormat.un8Sign == 0) {
        if(_tD.tFormat.un8Endian == 0) _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerU16m;  // U16 LE mono
        else                           _pfnDoSoundMixing = CAudioEnvCImp::_IDSPMixerU16m; // U16 BE mono
      }
      else {
        if(_tD.tFormat.un8Endian == 0) _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerS16m;  // S16 LE mono
        else                           _pfnDoSoundMixing = CAudioEnvCImp::_IDSPMixerS16m; // S16 BE mono
      }
    }
  }
  else {
    if(_tD.tFormat.un8Bits == 8) {
      if(_tD.tFormat.un8Sign == 0)     _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerU8s;   // U8 stereo
      else                             _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerS8s;   // S8 stereo
    }
    else {
      if(_tD.tFormat.un8Sign == 0) {
        if(_tD.tFormat.un8Endian == 0) _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerU16s;  // U16 LE stereo
        else                           _pfnDoSoundMixing = CAudioEnvCImp::_IDSPMixerU16s; // U16 BE stereo
      }
      else {
        if(_tD.tFormat.un8Endian == 0) _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerS16s;  // S16 LE stereo
        else                           _pfnDoSoundMixing = CAudioEnvCImp::_IDSPMixerS16s; // S16 BE stereo
      }
    }
  }

  // Show info
  VInfo(_NFL_, "Realized audio device settings :\n  Device name         = %s\n  Format              = %s\n  Fragment size       = %u bytes\n  Channels            = %d\n  Frequency           = %d Hz\n  Primary buffer size = %u",
        _tD.acDevName, VSoundBufferFormatName(_eFormat),
        _tD.un32FragSize, _tD.tFormat.un8Channels,
        _tD.tFormat.un32Frequency, _tD.un32PBuffSize);
  if(ptPreferred) {
    if(ptR->eFormat!=ptP->eFormat || ptR->n32Channels!=ptP->n32Channels || ptR->n32Frequency!=ptP->n32Frequency)
      VInfo(_NFL_, "The realized audio settings are different than the preferred settings");
  }
}

void CAlsaAudioEnv::ActivateDevice()
{
  // Check
  if(!_tD.deviceHandle)
    VError(VERR_DeviceNotOpenend, _NFL_, "Audio device is not opened");

  // Clear buffer
  _ClearBuffer_(_tD.pucPBuff, _tD.un32PBuffSize, _tD.tFormat.un8Sign, _tD.tFormat.un8Bits);

  // Activate writer thread
  _csTWriter.Start();
}

void CAlsaAudioEnv::DeactivateDevice()
{
  // Check
  if(!_tD.deviceHandle) return;

  // Deactivate writer thread
  try {
    _csTWriter.Stop();
  }
  catch(...) {
    _csTWriter.Kill();
  }
}

void CAlsaAudioEnv::GetAudioSpec(uint32_t *un32Channels, uint32_t *un32Bits, uint32_t *un32Rates, ESoundBufferFormat *eFormat, uint32_t *un32FragSize) const
{
  // Check
  if(!_tD.deviceHandle) {
    *un32Channels = 0;
    *un32Bits     = 0;
    *un32Rates    = 0;
    *eFormat      = ESB_Unknown;
    *un32FragSize = 0;
  }

  // Return the format
  *un32Channels = _tD.tFormat.un8Channels;
  *un32Bits     = _tD.tFormat.un8Bits;
  *un32Rates    = _tD.tFormat.un32Frequency;
  *eFormat      = _eFormat;
  *un32FragSize = _tD.un32FragSize;
}

void CAlsaAudioEnv::SetMidiPlayer(CMidiPlayer *pcsPlayer)
{
  _pcsMidiPlayer = pcsPlayer;
}

void CAlsaAudioEnv::_ClearBuffer_(u_char *pucBuff, uint32_t un32BuffSize, uint8_t un8Sign, uint8_t un8Bits)
{
  if(un8Sign)
    memset(pucBuff, 0x00, un32BuffSize);
  else if(un8Bits == 8)
    memset(pucBuff, 0x80, un32BuffSize);
  else {
    uint16_t *buf16 = reinterpret_cast<uint16_t*>(pucBuff);
    for(size_t i = 0; i < (un32BuffSize >> 1); ++i) *buf16++ = 0x8000;
  }
}

CSoundBuffer *CAlsaAudioEnv::NewSoundBuffer()
{
  return(_nothrow_new CAlsaSoundBuffer);
}

uint64_t CAlsaAudioEnv::_ThreadProc(TThreadData *ptThreadData)
{
  VInfo(_NFL_, "ALSA thread activated");

  while(!(*ptThreadData->pbExit)) {

    // Suspended ?
    _if_unlikely(*ptThreadData->pbSuspend) {
      usleep(100000);
      continue;
    }
    usleep(1000);

    // * If there is no available sound buffer data, don't write
    //   to the audio device.
    // * Using this mechanism, the display can be synced with the
    //   audio if the sound buffer updating function is called
    //   from the display (frame updating) loop.
    if(CAlsaAudioEnv::_instance->_tD.un8ToWrite > 3) continue;

    // Mix with the MIDI data
    if(CAlsaAudioEnv::_instance->_pcsMidiPlayer) {
       CAlsaAudioEnv::_instance->_pcsMidiPlayer->_MixWithBuffer(CAlsaAudioEnv::_instance->_tD.pucPBuffRead,
                                                                CAlsaAudioEnv::_instance->_tD.un32FragSize,
                                                                ptThreadData->pbExit);
    }

    // Write data
    int32_t sz = CAlsaAudioEnv::_instance->_tD.un32WriteSize;
    u_char *bf = CAlsaAudioEnv::_instance->_tD.pucPBuffRead;
    while(sz > 0) {
      int32_t ret = snd_pcm_writei(CAlsaAudioEnv::_instance->_tD.deviceHandle, bf, sz);
      if(ret < 0) {
        if(ret == -EAGAIN)
          continue;
        else if(ret == -EPIPE) {
          ret = snd_pcm_prepare(CAlsaAudioEnv::_instance->_tD.deviceHandle);
          if(ret < 0) {
            string msg = "Couldn't recover ALSA from buffer underrun ";
            msg += "("; msg += snd_strerror(ret); msg += ")";
            VAbortOnFatalError(true);
            VFatalError(VERR_FATAL_IOError, _NFL_, msg.c_str());
          }
        }
        else if(ret == -ESTRPIPE) {
          while((ret = snd_pcm_resume(CAlsaAudioEnv::_instance->_tD.deviceHandle)) == -EAGAIN) usleep(1000);
          if(ret < 0) ret = snd_pcm_prepare(CAlsaAudioEnv::_instance->_tD.deviceHandle);
          if(ret < 0) {
            string msg = "Couldn't recover ALSA from suspend ";
            msg += "("; msg += snd_strerror(ret); msg += ")";
            VAbortOnFatalError(true);
            VFatalError(VERR_FATAL_IOError, _NFL_, msg.c_str());
          }
        }
      }
      sz -= ret;
      bf += (ret * CAlsaAudioEnv::_instance->_tD.un32WriteIncF);
    }

    // Increment the "ToWrite" value
    CAlsaAudioEnv::_instance->_csMutex.Lock();
    CAlsaAudioEnv::_instance->_tD.un8ToWrite++;
    CAlsaAudioEnv::_instance->_csMutex.Unlock();

    // Increment the read pointer
    CAlsaAudioEnv::_instance->_tD.pucPBuffRead += CAlsaAudioEnv::_instance->_tD.un32FragSize;
    if(CAlsaAudioEnv::_instance->_tD.pucPBuffRead > CAlsaAudioEnv::_instance->_tD.pucPBuffMax) CAlsaAudioEnv::_instance->_tD.pucPBuffRead = CAlsaAudioEnv::_instance->_tD.pucPBuff;
  }

  VInfo(_NFL_, "ALSA thread deactivated");
  return(0);
}

} // namespace vge
#endif
