/** \file MediaPool.tpp.h */
// ############################################################################################
//  MediaPool.tpp.h
//  Implementation of CBasicMediaPool template class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _GAME_OBJECT___MEDIA_POOL_TPP_H_
#define _GAME_OBJECT___MEDIA_POOL_TPP_H_

// VortexGE headers
#include "MediaPool.h"


namespace vge {

//
// _CMediaInstE implementation
//
template<typename MediaT>
MediaT *_CMediaInstE<MediaT>::Get()
{ 
  vassert(_mediaInst, "Couldn't Get() a NULL media resource");

  _lastGetTime = time(NULL); 
  _refCount++; 
  return(_mediaInst);
}

template<typename MediaT>
void _CMediaInstE<MediaT>::Release()
{
  vassert(_refCount > 0,
          "Unbalance Get() and Release() of media resource %p",
          static_cast<const void*>(_mediaInst));

  _refCount--; 
}


//
// CBasicMediaPool implementation
//
template<typename TResourceID>
CBasicMediaPool<TResourceID>::CBasicMediaPool(CAudioEnv *audioEnvironment, bool throwExceptionIfResourceNotFound)
: _audEnv(audioEnvironment), _throwIfResNotFound(throwExceptionIfResourceNotFound)
{
}

template<typename TResourceID>
CBasicMediaPool<TResourceID>::~CBasicMediaPool() throw()
{
  DeleteAllImages(true);
  DeleteAllAnimatedImages(true);
  DeleteAllTextures(true);
  DeleteAllAnimatedTextures(true);
  DeleteAllSoundBuffers(true);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::SetAudioEnvironment(CAudioEnv *audioEnvironment)
{
  _audEnv = audioEnvironment;
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::ThrowExceptionIfResourceNotFound(bool throws)
{
  _throwIfResNotFound = throws;
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::Compact(time_t timeLimitInSeconds)
{
  _Compact(_imageMap, timeLimitInSeconds);
  _Compact(_aimageMap, timeLimitInSeconds);
  _Compact(_textureMap, timeLimitInSeconds);
  _Compact(_atextureMap, timeLimitInSeconds);
  _Compact(_soundbufferMap, timeLimitInSeconds);
}


//
// CBasicMediaPool implementation : helper functions
//
template<typename TResourceID>
template<typename MapT>
void CBasicMediaPool<TResourceID>::_Compact(MapT &map, time_t timeLimitInSeconds)
{
  // Proceed only if the map is not empty
  if(map.empty()) return;

  // For convenience
  typedef typename MapT::iterator IterT;

  // Calculate the time
  const time_t currentTime = time(NULL);

  // Our "to be keep" and "to be deleted" list
  MapT _keep, _del;

  // Search trough the map and find which instance that can be deleted
  for(IterT it = map.begin(); it != map.end(); ++it) {
    // If it is still referenced by anything, it can't be deleted
    if(it->second.IsReferenced())
      _keep[it->first] = it->second;
    // If it is no longer referenced by anything, check if it is
    // still recent enough thus it will not be deleted
    else if((currentTime - it->second.LastGetTime()) < timeLimitInSeconds)
      _keep[it->first] = it->second;
    // It is no longer referenced by anything and also already
    // old enough to be deleted
    else
      _del[it->first] = it->second;
  }

  // Proceed only if "to be deleted" is not empty
  if(!_del.empty()) {
    // Swap the "to be keep" map with the original map
    swap(map, _keep);
    // Delete the old instances
    for(IterT it = _del.begin(); it != _del.end(); ++it) {
      it->second.MediaInst()->MakeItAbortOnDestruction(false);
      delete it->second.MediaInst();
    }
  }
}

template<typename TResourceID>
template<typename MapT>
void CBasicMediaPool<TResourceID>::_ThrowIfMediaAlreadyExist(const MapT &map, const TResourceID &id, const char* mname)
{
  if(map.find(id) != map.end())
    VError(VERR_InvalidParameter, _NFL_, "Duplicated resource ID for \"%s\" resource", mname);
}

template<typename TResourceID>
template<typename EntryT, typename MapT>
void CBasicMediaPool<TResourceID>::_AddMediaToList(MapT &map, const TResourceID &id, typename MapT::mapped_type::MediaT *media)
{
  map[id] = EntryT(media);
  media->MakeItAbortOnDestruction(true);
}

template<typename TResourceID>
template<typename MapT>
bool CBasicMediaPool<TResourceID>::_IsRegisteredMedia(const MapT &map, const TResourceID &id)
{
  return(map.find(id) != map.end());
}

template<typename TResourceID>
template<typename MapT>
typename MapT::mapped_type::MediaT *CBasicMediaPool<TResourceID>::_GetMedia(MapT &map, const TResourceID &id, const char* mname, bool throwIfNotExist)
{
  typename MapT::iterator it = map.find(id);
  if(it == map.end()) {
    if(throwIfNotExist) VError(VERR_NotExist, _NFL_,
                               "Requested resource \"%s\" is not exist", mname);
    return(NULL);
  }
  return(it->second.Get());
}

template<typename TResourceID>
template<typename MapT>
typename MapT::mapped_type::MediaT *CBasicMediaPool<TResourceID>::_GetMediaInst(MapT &map, const typename MapT::mapped_type::MediaT *inst, const char* mname, bool throwIfNotExist)
{
  for(typename MapT::iterator it = map.begin(); it != map.end(); ++ it) {
    if(it->second.MediaInst() == inst) {
      return(it->second.Get());
    }
  }
  if(throwIfNotExist) VError(VERR_NotExist, _NFL_,
                             "Requested instance %p of resource \"%s\" is not exist", inst, mname);
  return(NULL);
}

template<typename TResourceID>
template<typename MapT>
void CBasicMediaPool<TResourceID>::_ReleaseMedia(MapT &map, const TResourceID &id, const char* mname, bool throwIfNotExist)
{
  typename MapT::iterator it = map.find(id);
  if(it == map.end()) {
    if(throwIfNotExist) 
      if(throwIfNotExist) VError(VERR_NotExist, _NFL_,
                                 "Requested resource \"%s\" is not exist", mname);
  }
  it->second.Release();
}

template<typename TResourceID>
template<typename MapT>
void CBasicMediaPool<TResourceID>::_ReleaseMediaInst(MapT &map, const typename MapT::mapped_type::MediaT *inst, const char* mname, bool throwIfNotExist)
{
  for(typename MapT::iterator it = map.begin(); it != map.end(); ++ it) {
    if(it->second.MediaInst() == inst) {
      it->second.Release();
      return;
    }
  }
  if(throwIfNotExist) VError(VERR_NotExist, _NFL_,
                             "Requested instance %p of resource \"%s\" is not exist", inst, mname);
}

template<typename TResourceID>
template<typename MapT>
void CBasicMediaPool<TResourceID>::_DeleteMedia(MapT &map, const TResourceID &id, const char* mname, bool forceEvenStillUsed)
{
  typename MapT::iterator it = map.find(id);
  if(it == map.end()) return;

  if(!forceEvenStillUsed && it->second.IsReferenced())
    VError(VERR_InvalidCall, _NFL_, "Couldn't delete the still referenced \"%s\" resource %p", mname, static_cast<const void*>(it->second.MediaInst()));

  it->second.MediaInst()->MakeItAbortOnDestruction(false);
  delete it->second.MediaInst();
  map.erase(it);
}

template<typename TResourceID>
template<typename MapT>
void CBasicMediaPool<TResourceID>::_DeleteAllMedias(MapT &map, const char* mname, bool forceEvenStillUsed)
{
  typedef typename MapT::iterator       IterT;
  typedef typename MapT::const_iterator ConstIterT;

  // Check if any of the resources is still used
  if(!forceEvenStillUsed) {
    for(ConstIterT it = map.begin(); it != map.end(); ++it) {
      if(it->second.IsReferenced())
        VError(VERR_InvalidCall, _NFL_, "Couldn't delete the still referenced \"%s\" resource %p", mname, static_cast<const void*>(it->second.MediaInst()));
    }
  }

  // Delete all the resources
  for(IterT it = map.begin(); it != map.end(); ++it) {
    it->second.MediaInst()->MakeItAbortOnDestruction(false);
    delete it->second.MediaInst();
  }
  map.clear();
}


//
// CBasicMediaPool implementation : CImage related functions
//
template<typename TResourceID>
void CBasicMediaPool<TResourceID>::NewImage(const TResourceID &id, const CImageBuilder &src)
{
  _ThrowIfMediaAlreadyExist(_imageMap, id, "CImage");

  CImage *nm = new CImage;
  src.TransferToImage(*nm);

  _AddMediaToList<_CImageE>(_imageMap, id, nm);
}

template<typename TResourceID>
bool CBasicMediaPool<TResourceID>::IsRegisteredImage(const TResourceID &id) const
{
  return(_IsRegisteredMedia(_imageMap, id));
}

template<typename TResourceID>
CImage *CBasicMediaPool<TResourceID>::GetImage(const TResourceID &id)
{
  return(_GetMedia(_imageMap, id, "CImage", _throwIfResNotFound));
}

template<typename TResourceID>
CImage *CBasicMediaPool<TResourceID>::GetImage(const CImage *inst)
{
  return(_GetMediaInst(_imageMap, inst, "CImage", _throwIfResNotFound));
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::ReleaseImage(const TResourceID &id)
{
  _ReleaseMedia(_imageMap, id, "CImage", _throwIfResNotFound);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::ReleaseImage(const CImage *inst)
{
  _ReleaseMediaInst(_imageMap, inst, "CImage", _throwIfResNotFound);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::DeleteImage(const TResourceID &id, bool forceEvenStillUsed)
{
  _DeleteMedia(_imageMap, id, "CImage", forceEvenStillUsed);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::DeleteAllImages(bool forceEvenStillUsed)
{
  _DeleteAllMedias(_imageMap, "CImage", forceEvenStillUsed);
}


//
// CBasicMediaPool implementation : CAnimatedImage related functions
//
template<typename TResourceID>
void CBasicMediaPool<TResourceID>::NewAnimatedImage(const TResourceID &id, const CImageBuilder &src, uint8_t rowCount, uint8_t colCount)
{
  _ThrowIfMediaAlreadyExist(_aimageMap, id, "CAnimatedImage");

  CAnimatedImage *nm = new CAnimatedImage;
  src.TransferToAnimatedImage(*nm, rowCount, colCount);

  _AddMediaToList<_CAnimatedImageE>(_aimageMap, id, nm);
}

template<typename TResourceID>
bool CBasicMediaPool<TResourceID>::IsRegisteredAnimatedImage(const TResourceID &id) const
{
  return(_IsRegisteredMedia(_aimageMap, id));
}

template<typename TResourceID>
CAnimatedImage *CBasicMediaPool<TResourceID>::GetAnimatedImage(const TResourceID &id)
{
  return(_GetMedia(_aimageMap, id, "CAnimatedImage", _throwIfResNotFound));
}

template<typename TResourceID>
CAnimatedImage *CBasicMediaPool<TResourceID>::GetAnimatedImage(const CAnimatedImage *inst)
{
  return(_GetMediaInst(_aimageMap, inst, "CAnimatedImage", _throwIfResNotFound));
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::ReleaseAnimatedImage(const TResourceID &id)
{
  _ReleaseMedia(_aimageMap, id, "CAnimatedImage", _throwIfResNotFound);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::ReleaseAnimatedImage(const CAnimatedImage *inst)
{
  _ReleaseMediaInst(_aimageMap, inst, "CAnimatedImage", _throwIfResNotFound);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::DeleteAnimatedImage(const TResourceID &id, bool forceEvenStillUsed)
{
  _DeleteMedia(_aimageMap, id, "CAnimatedImage", forceEvenStillUsed);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::DeleteAllAnimatedImages(bool forceEvenStillUsed)
{
  _DeleteAllMedias(_aimageMap, "CAnimatedImage", forceEvenStillUsed);
}


//
// CBasicMediaPool implementation : CTexture related functions
//
template<typename TResourceID>
void CBasicMediaPool<TResourceID>::NewTexture(const TResourceID &id, const CImageBuilder &src)
{
  _ThrowIfMediaAlreadyExist(_textureMap, id, "CTexture");

  CTexture *nm = new CTexture;
  src.TransferToTexture(*nm);

  _AddMediaToList<_CTextureE>(_textureMap, id, nm);
}

template<typename TResourceID>
bool CBasicMediaPool<TResourceID>::IsRegisteredTexture(const TResourceID &id) const
{
  return(_IsRegisteredMedia(_textureMap, id));
}

template<typename TResourceID>
CTexture *CBasicMediaPool<TResourceID>::GetTexture(const TResourceID &id)
{
  return(_GetMedia(_textureMap, id, "CTexture", _throwIfResNotFound));
}

template<typename TResourceID>
CTexture *CBasicMediaPool<TResourceID>::GetTexture(const CTexture *inst)
{
  return(_GetMediaInst(_textureMap, inst, "CTexture", _throwIfResNotFound));
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::ReleaseTexture(const TResourceID &id)
{
  _ReleaseMedia(_textureMap, id, "CTexture", _throwIfResNotFound);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::ReleaseTexture(const CTexture *inst)
{
  _ReleaseMediaInst(_textureMap, inst, "CTexture", _throwIfResNotFound);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::DeleteTexture(const TResourceID &id, bool forceEvenStillUsed)
{
  _DeleteMedia(_textureMap, id, "CTexture", forceEvenStillUsed);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::DeleteAllTextures(bool forceEvenStillUsed)
{
  _DeleteAllMedias(_textureMap, "CTexture", forceEvenStillUsed);
}


//
// CBasicMediaPool implementation : CAnimatedTexture related functions
//
template<typename TResourceID>
void CBasicMediaPool<TResourceID>::NewAnimatedTexture(const TResourceID &id, const CImageBuilder &src, uint8_t rowCount, uint8_t colCount)
{
  _ThrowIfMediaAlreadyExist(_atextureMap, id, "CAnimatedTexture");

  CAnimatedTexture *nm = new CAnimatedTexture;
  src.TransferToAnimatedTexture(*nm, rowCount, colCount);

  _AddMediaToList<_CAnimatedTextureE>(_atextureMap, id, nm);
}

template<typename TResourceID>
bool CBasicMediaPool<TResourceID>::IsRegisteredAnimatedTexture(const TResourceID &id) const
{
  return(_IsRegisteredMedia(_atextureMap, id));
}

template<typename TResourceID>
CAnimatedTexture *CBasicMediaPool<TResourceID>::GetAnimatedTexture(const TResourceID &id)
{
  return(_GetMedia(_atextureMap, id, "CAnimatedTexture", _throwIfResNotFound));
}

template<typename TResourceID>
CAnimatedTexture *CBasicMediaPool<TResourceID>::GetAnimatedTexture(const CAnimatedTexture *inst)
{
  return(_GetMediaInst(_atextureMap, inst, "CAnimatedTexture", _throwIfResNotFound));
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::ReleaseAnimatedTexture(const TResourceID &id)
{
  _ReleaseMedia(_atextureMap, id, "CAnimatedTexture", _throwIfResNotFound);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::ReleaseAnimatedTexture(const CAnimatedTexture *inst)
{
  _ReleaseMediaInst(_atextureMap, inst, "CAnimatedTexture", _throwIfResNotFound);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::DeleteAnimatedTexture(const TResourceID &id, bool forceEvenStillUsed)
{
  _DeleteMedia(_atextureMap, id, "CAnimatedTexture", forceEvenStillUsed);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::DeleteAllAnimatedTextures(bool forceEvenStillUsed)
{
  _DeleteAllMedias(_atextureMap, "CAnimatedTexture", forceEvenStillUsed);
}


//
// CBasicMediaPool implementation : CSoundBuffer related functions
//
template<typename TResourceID>
void CBasicMediaPool<TResourceID>::NewSoundBuffer(const TResourceID      &id,
                                                  const CDataSource      &srcAudioData,
                                                  ESourceAudioDataFormat  sourceAudioDataFormat,
                                                  ERawFormat              rawDataFormat,
                                                  int32_t                 rawDataFrequency)
{
  // Chek if audio is disabled at initialization
  if(!_audEnv)
    VError(VERR_NotInitialized, _NFL_, "Audio support is not enabled when initializing this CBasicMediaPool instance");

  // Check for duplicated resource ID
  _ThrowIfMediaAlreadyExist(_soundbufferMap, id, "CSoundBuffer");

  // Determine the source data format
  if(sourceAudioDataFormat == ESADF_AUTO) {
    char pb[4];
    srcAudioData.PStream().peek(pb, 4);
         if(strncmp(pb, "VSW",  3) == 0) sourceAudioDataFormat = ESADF_VSW;
    else if(strncmp(pb, "RIFF", 4) == 0) sourceAudioDataFormat = ESADF_WAV;
    else                                 sourceAudioDataFormat = ESADF_RAW;
  }

  // Check the parameters
  if(sourceAudioDataFormat == ESADF_RAW) {
    if(rawDataFormat==ERF_Invalid || !rawDataFrequency)
    VError(VERR_InvalidParameter, _NFL_, "The \"rawDataFormat\" and/or \"rawDataFrequency\" parameter(s) contain(s) (an) invalid value(s)\nMaybe you are not actually intended to load the data as raw PCM ?");
  }

  // Create a new sound buffer
  CSoundBuffer *nm = _audEnv->NewSoundBuffer();

  // Load the audio data
  switch(sourceAudioDataFormat) {
    case ESADF_RAW: nm->LoadRAW(srcAudioData, rawDataFormat, rawDataFrequency); break;
    case ESADF_VSW: nm->LoadVSW(srcAudioData); break;
    case ESADF_WAV: nm->LoadWAV(srcAudioData); break;
    default       : delete nm; VError(VERR_InvalidParameter, _NFL_, "Invalid source-audio data-format constant");
  }

  // Put it in the list
  _AddMediaToList<_CSoundBufferE>(_soundbufferMap, id, nm);
}

template<typename TResourceID>
bool CBasicMediaPool<TResourceID>::IsRegisteredSoundBuffer(const TResourceID &id) const
{
  return(_IsRegisteredMedia(_soundbufferMap, id));
}

template<typename TResourceID>
CSoundBuffer *CBasicMediaPool<TResourceID>::GetSoundBuffer(const TResourceID &id)
{
  return(_GetMedia(_soundbufferMap, id, "CSoundBuffer", _throwIfResNotFound));
}

template<typename TResourceID>
CSoundBuffer *CBasicMediaPool<TResourceID>::GetSoundBuffer(const CSoundBuffer *inst)
{
  return(_GetMediaInst(_soundbufferMap, inst, "CSoundBuffer", _throwIfResNotFound));
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::ReleaseSoundBuffer(const TResourceID &id)
{
  _ReleaseMedia(_soundbufferMap, id, "CSoundBuffer", _throwIfResNotFound);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::ReleaseSoundBuffer(const CSoundBuffer *inst)
{
  _ReleaseMediaInst(_soundbufferMap, inst, "CSoundBuffer", _throwIfResNotFound);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::DeleteSoundBuffer(const TResourceID &id, bool forceEvenStillUsed)
{
  _DeleteMedia(_soundbufferMap, id, "CSoundBuffer", forceEvenStillUsed);
}

template<typename TResourceID>
void CBasicMediaPool<TResourceID>::DeleteAllSoundBuffers(bool forceEvenStillUsed)
{
  _DeleteAllMedias(_soundbufferMap, "CSoundBuffer", forceEvenStillUsed);
}

} // namespace vge
#endif
