/** \file Plane3DBuilder.h */
// ############################################################################################
//  Entities.h
//  Definition & implementation of CPlane3DBuilder and CPlane3D classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _GAME_OBJECT___PLANE_3D_BUILDER_H_
#define _GAME_OBJECT___PLANE_3D_BUILDER_H_

// VortexGE headers
#include "../3DEnvironment/Panel3DBuffer.h"
#include "../Base/Random.h"


namespace vge {

    //!
    //! \brief Plane direction for plane generation
    //!
    enum EPlaneDir {
      EPD_0P0, //!< Normal( 0, 1, 0) -> as if it is the top side of a 3D cube
      EPD_0N0, //!< Normal( 0,-1, 0) -> as if it is the bottom side of a 3D cube
      EPD_P00, //!< Normal( 1, 0, 0) -> as if it is the right side of a 3D cube
      EPD_N00, //!< Normal(-1, 0, 0) -> as if it is the left side of a 3D cube
      EPD_00P, //!< Normal( 0, 0, 1) -> as if it is the back side of a 3D cube
      EPD_00N  //!< Normal( 0, 0,-1) -> as if it is the front side of a 3D cube
    };


  //!
  //! \brief A 3D plane class
  //!
  //! Shouldn't need to instantiate this class directly. Use it from
  //! within the CPlane3DBuilder class instead.
  //!
  //! \api beta
  class VGE_EXPORT CPlane3D {
    public:
      //! The default grid size is (100, 100, 100)
      static const TFloatXYZ DEFAULT_GRID_SIZE;

    public:
      //!
      //! Construct a new CPlane3D instance with the given grid size
      //! and plane origin.
      inline CPlane3D(const TFloatXYZ &gridSize    = DEFAULT_GRID_SIZE,
                      const TFloatXYZ &planeOrigin = TFloatXYZ(0, 0, 0))
      : _gridSize(gridSize), _planeOrigin(planeOrigin)
      {}
      //!
      //! Copy ctor
      inline CPlane3D(const CPlane3D &src)
      { *this = src; }
  
      // Initialization functions
      //!
      //! Set the grid size.
      inline void SetGridSize(const TFloatXYZ &newGridSize)
      { _gridSize = newGridSize; }
      //!
      //! Set the plane origin.
      inline void SetPlaneOrigin(const TFloatXYZ &newPlaneOrigin, bool updateExistingPlane = false)
      { _ChangePlaneOrigin(newPlaneOrigin, updateExistingPlane); }
  
      // Plane generation functions
      //!
      //! Generate plane with the given size and direction.
      inline void GeneratePlane(uint32_t planeRowSize, uint32_t planeColSize, EPlaneDir planeDir)
      { _GeneratePlane(planeRowSize, planeColSize, planeDir); }
      //!
      //! Translate the vertices' coordinates by the given factor.
      void TranslateVertices(const TFloatXYZ &factor);
      //!
      //! Randomly vary the plane's vertices heights (one of the vertices' X/Y/Z
      //! components which are in the direction with the plane's normal).
      //! \n\n
      //! <b>Use this function with care, since the bigger the random factor, the
      //! bigger the chance that this function will produce concave, twisted,
      //! squeezed, and/or any other types of "bad" faces that when rendered
      //! will cause rendering artifacts or worse they may crash the application.</b>
      void RandomlyVaryHeights(float              maxFactor,
                               bool               inProcentageOfGridSize = true,
                               bool               excludeOutlineVertices = true,
                               CRandomCoreUInt32 *randomCore             = NULL);

      // Information functions
      //!
      //! Returns the row size
      inline uint32_t GetRowSize() const
      { return(_planeRowSize); }
      //!
      //! Returns the column size
      inline uint32_t GetColSize() const
      { return(_planeColSize); }
      //!
      //! Returns the direction.
      inline EPlaneDir GetDirection() const
      { return(_planeDir); }
      //!
      //! Returns the number of genereated vertices.
      inline uint32_t GetNumberOfVertices() const
      { return(_vertices.size() / 3); }
      //!
      //! Returns the number of genereated faces.
      inline uint32_t GetNumberOfFaces() const
      { return(_faces.size() / 4); }
      //!
      //! Get the current grid size.
      inline const TFloatXYZ &GetGridSize() const
      { return(_gridSize); }
      //!
      //! Get the plane's bounding box.
      void GetBoundingBox(TFloatXYZ &minCoord, TFloatXYZ &maxCoord) const;
      //!
      //! Get the plane's origin.
      void GetOriginPosition(TFloatXYZ &orgCoord) const;
  
      // Direct access to the plane's data
      //!
      //! Direct access to the plane's vertices.
      inline vector<float> &GetVertices()
      { return(_vertices); }
      //!
      //! Direct access to the plane's vertices.
      inline const vector<float> &GetVertices() const
      { return(_vertices); }
      //!
      //! Direct access to the plane's vertices raw data buffer.
      inline const float *GetRawVerticesBuffer() const
      { return(&_vertices[0]); }
      //!
      //! Direct access to the plane's faces.
      inline vector<uint32_t> &GetFaces()
      { return(_faces); }
      //!
      //! Direct access to the plane's faces.
      inline const vector<uint32_t> &GetFaces() const
      { return(_faces); }
      //!
      //! Direct access to the plane's faces raw data buffer.
      inline const uint32_t *GetRawFacesBuffer() const
      { return(&_faces[0]); }

      // Overloaded operators
      CPlane3D &operator=(const CPlane3D &src);

      // I/O functions
      //!
      //! Load contents from a pistream.
      void LoadFromStream(pistream &is);
      //!
      //! Save contents to an ostream.
      void SaveToStream(ostream &os) const;

    protected:
      // Helper functions
      void _GeneratePlane(uint32_t planeRowSize, uint32_t planeColSize, EPlaneDir planeDir);
      void _ChangePlaneOrigin(const TFloatXYZ &newPlaneOrigin, bool updateExistingPlane);
  
      // Data
      vector<float>    _vertices;     // Vertices buffer
      vector<uint32_t> _faces;        // Faces buffer
      TFloatXYZ        _gridSize;     // Grid size to be used when generating the plane
      TFloatXYZ        _planeOrigin;  // Origin of the plane
      uint32_t         _planeRowSize; // Total rows of the plane
      uint32_t         _planeColSize; // Total columns of the plane
      EPlaneDir        _planeDir;     // Direction of the plane
  };


  //!
  //! \brief A 3D plane builder class
  //!
  //! \api experiment
  class VGE_EXPORT CPlane3DBuilder : VCLASS_UNCOPYABLE {
    public:
      //! Map of indentifiers and the corresponding CPlane3D instances.
      typedef map<string, CPlane3D*> CPlane3DMap;

      //! The default grid size is (100, 100, 100).
      static const TFloatXYZ &DEFAULT_GRID_SIZE;

      //! The default error limit is 0.001f.
      static const float DEFAULT_ERROR_LIMIT;

    public:
      //!
      //! Construct a new CPlane3DBuilder instance with the given grid size.
      inline CPlane3DBuilder(const TFloatXYZ &gridSize = DEFAULT_GRID_SIZE)
      : _gridSize(gridSize)
      {}
      //!
      //! Standard dtor.
      ~CPlane3DBuilder() throw();
  
      // Initialization functions
      //!
      //! Set the grid size.
      inline void SetGridSize(const TFloatXYZ &newGridSize = DEFAULT_GRID_SIZE)
      { _gridSize = newGridSize; }
      //!
      //! Clear (delete) all genereated planes.
      void ClearAll();

      // Plane generation functions
      //!
      //! Generate plane with the given size and direction.
      void GeneratePlane(const string    &planeName,
                         const TFloatXYZ &planeOrigin,
                         uint32_t         planeRowSize,
                         uint32_t         planeColSize, 
                         EPlaneDir        planeDir);
      //!
      //! Rename plane with the given name with the given new name.
      void RenamePlane(const string &oldPlaneName, const string &newPlaneName);
      //!
      //! Delete plane with the given name, will ignore if not exist.
      void DeletePlane(const string &planeName);
      //!
      //! Delete planes with the given names, will ignore if not exist.
      void DeletePlane(const vector<string> &planeNames);
      //!
      //! Translate all the planes' vertices' coordinates by the given factor.
      void TranslateVertices(const TFloatXYZ &factor);
      //!
      //! Translate all the selected planes' vertices' coordinates by the given factor.
      void TranslateVertices(const TFloatXYZ      &factor,
                             const vector<string> &selectedPlanesName,
                             bool                  ignoreIfNotExist = false);
      //!
      //! Combine all the copies of the planes in the given CPlane3DBuilder
      //! instance with this CPlane3DBuilder instance.
      //! \param[in] src        CPlane3DBuilder instance of which planes will be 
      //!                       copied from.
      //! \param[in] namePrefix Optional string to be appended to the generated
      //!                       planes' name. The function will appends a '-' at
      //!                       the end of the name string if the name is not an
      //!                       empty string.
      void Combine(const CPlane3DBuilder &src, const string &namePrefix = "");
      //!
      //! Combine all the copies of the selected planes in the given CPlane3DBuilder
      //! instance with this CPlane3DBuilder instance.
      //! \param[in] src              CPlane3DBuilder instance of which planes will be 
      //!                             copied from.
      //! \param[in] srcPlaneNames    Array containing the name of the source planes to
      //!                             be combined.
      //! \param[in] namePrefix       Optional string to be appended to the generated
      //!                             planes' name. The function will appends a '-' at
      //!                             the end of the name string if the name is not an
      //!                             empty string.
      //! \param[in] ignoreIfNotExist If "true", just ignore if srcPlaneNames contains
      //!                             non existing source plane's names.
      void Combine(const                 CPlane3DBuilder &src, 
                   const vector<string> &srcPlaneNames,
                   const string         &namePrefix       = "",
                   bool                  ignoreIfNotExist = false);
      //!
      //! Fills the given CPanel3DBuffer instance with all the generated planes.
      void GeneratePanel3DBuffer(CPanel3DBuffer &dstPanel3DBuffer,
                                 float           combineVertexErrorLimit = DEFAULT_ERROR_LIMIT) const;
      //!
      //! Fills the given CPanel3DBuffer instance with all the generated
      //! planes, apply the given texture to it, and finally, generate
      //! default texture coordinates for the CPanel3DBuffer instance.
      void GeneratePanel3DBuffer(CPanel3DBuffer &dstPanel3DBuffer,
                                 CTexture       *textureToBeUsed,
                                 float           combineVertexErrorLimit = DEFAULT_ERROR_LIMIT) const;
      //!
      //! Fills the given CPanel3DBuffer instance with all the selected planes.
      void GeneratePanel3DBuffer(CPanel3DBuffer       &dstPanel3DBuffer,
                                 const vector<string> &selectedPlanesName,
                                 float                 combineVertexErrorLimit = DEFAULT_ERROR_LIMIT,
                                 bool                  ignoreIfNotExist        = false) const;
      //!
      //! Fills the given CPanel3DBuffer instance with all the selected
      //! planes, apply the given texture to it, and finally, generate
      //! default texture coordinates for the CPanel3DBuffer instance.
      void GeneratePanel3DBuffer(CPanel3DBuffer       &dstPanel3DBuffer,
                                 const vector<string> &selectedPlanesName,
                                 CTexture             *textureToBeUsed,
                                 float                 combineVertexErrorLimit = DEFAULT_ERROR_LIMIT,
                                 bool                  ignoreIfNotExist        = false) const;

      // Information functions
      //!
      //! Get the current grid size.
      inline const TFloatXYZ &GetGridSize() const
      { return(_gridSize); }
      //!
      //! Get the bounding box of all planes
      void GetBoundingBox(TFloatXYZ &minCoord, TFloatXYZ &maxCoord) const;
      //!
      //! Get the name of all the planes
      void GetPlaneNames(vector<string> dstBuff) const;
  
      // Direct access to the map of planes.
      //!
      //! Direct access to the map of planes.
      inline CPlane3DMap &GetPlanes()
      { return(_planes); }
      //!
      //! Direct access to the map of planes.
      const inline CPlane3DMap &GetPlanes() const
      { return(_planes); }
      //!
      //! Direct access to a particular plane.
      CPlane3D &operator [] (const string &planeName);
      //!
      //! Direct access to a particular plane.
      const CPlane3D &operator [] (const string &planeName) const;

      // I/O functions
      //!
      //! Save to VortexGE's native VPB format.
      //! \param[in] sink The name of the file (.vpb) to be saved to or an instance of ostream derived class.
      void SaveVPB(const CDataSink &sink) const;
      //!
      //! Load from VortexGE's native VPB format.
      //! \param[in] source The name of the file (.vpb) to be loaded or an instance of istream derived class.
      void LoadVPB(const CDataSource &source);

      // Higher level construction functions
      //!
      //! Create simple open room.
      //! \param[in] roomOrigin           See the picture below.
      //! \param[in] roomSize             See the picture below.
      //! \param[in] leftWallThickness    See the picture below.
      //! \param[in] rightWallThickness   See the picture below.
      //! \param[in] backWallThickness    See the picture below.
      //! \param[in] floorThickness       See the picture below.
      //! \param[in] fullWall             If "true", additional walls will be generated.
      //! \param[in] namePrefix           Optional string to be appended to the
      //!                                 generated planes' name. The function will
      //!                                 appends a '-' at the end of the name string
      //!                                 if the name is not an empty string.
      //! \param[in] deleteExistingPlanes If "true", all previously created planes
      //!                                 will be deleted first.
      //!
      //! <img src="../../images/sora.png">
      void CreateSimpleOpenRoom(const TFloatXYZ &roomOrigin,
                                const TInt32XYZ &roomSize,
                                int32_t         leftWallThickness    = 1,
                                int32_t         rightWallThickness   = 1,
                                int32_t         backWallThickness    = 1,
                                int32_t         floorThickness       = 0,
                                bool            fullWall             = false,
                                const string   &namePrefix           = "",
                                bool            deleteExistingPlanes = false);

    protected:
      // Helper functions
      void _CombinePlanes(vector<float>          &verticesOut,
                          vector<uint32_t>       &facesOut,
                          const vector<float>    &verticesIn,
                          const vector<uint32_t> &facesIn,
                          float                   errLimit) const;
  
      // Data
      CPlane3DMap _planes;   // Map of planes
      TFloatXYZ   _gridSize; // Grid size to be used when generating the planes
  };

} // namespace vge
#endif
