// ############################################################################################
//  FrameRateControl.cc
//  Implementation of CFPSControl class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../VideoEnvironment/VideoEnvironment.h"
#include "FrameRateControl.h"


namespace vge {

//
// CFPSControl implementation
//
CFPSControl::CFPSControl(float fWantedFPS)
: _un64STicks(0),          _un64ETicks(0),
  _un64Frames(0),          _n32Frames (0),
  _n32Dly    (0),          _n32DlyMax (0),
  _n32FPSD   (0),
  _tTicks1   (0),          _tTicks2   (0),
  _fFPS      (fWantedFPS), _fWantedFPS(fWantedFPS)
{}

float CFPSControl::CalculateFPS()
{
  _n32Frames++;
  _tTicks2 = VGetSystemTicks();

  uint64_t un64Dif = _tTicks2 - _tTicks1;
  if(un64Dif >= TICKS_PER_SEC) {
    _fFPS      = float(_n32Frames) / float(un64Dif) * float(TICKS_PER_SEC);
    _tTicks1   = _tTicks2;
    _n32Frames = 0;
  }

  return(_fFPS);
}

bool CFPSControl::PumpEventAndDelay(CVideoEnv &csVidEnv, bool bDisableDelay)
{
  // Frame rate control disabled
  _if_unlikely(bDisableDelay) {
    _n32Dly = 0;
    _if_unlikely(!csVidEnv.IsActive()) {
      if(csVidEnv.AnyPendingEvent()) return(false);
      csVidEnv.BlockedDispatchEvent();
      if(csVidEnv.AnyPendingEvent()) return(false);
      return(true);
    }
    else {
      csVidEnv.DispatchEvent();
      return(false);
    }
  }
  // Frame rate control enabled and the window is inactive
  // ---> no need for frame rate control
  else if(!csVidEnv.IsActive()) {
    if(csVidEnv.AnyPendingEvent()) return(false);
    csVidEnv.BlockedDispatchEvent();
    if(csVidEnv.AnyPendingEvent()) return(false);
  }
  // Frame rate control enabled and the window is active
  else {
    if(_n32Dly <= _n32DlyMax) {
      _n32Dly++;
      csVidEnv.DispatchEvent();
      if(csVidEnv.AnyPendingEvent()) return(false);
    }
    else {
      _n32Dly = 0;
      return(false);
    }
  }
  return(true);
}

bool CFPSControl::AdjustDelay(bool bDisableDelay)
{
  _if_unlikely(bDisableDelay) {
    _un64Frames++;
    _n32DlyMax = 0;
    CalculateFPS();
    return(true);
  }
  else {
    if(_n32Dly == 0) {
      _un64Frames++;
      CalculateFPS();
      _n32FPSD = int32_t(_fFPS - _fWantedFPS);
      if(_n32FPSD < 0) {
        _n32DlyMax += _n32FPSD;
        _n32DlyMax = (_n32DlyMax < 0) ? 0 : _n32DlyMax;
      }
      else if(_n32FPSD > 0) {
        _n32DlyMax += _n32FPSD;
        _n32DlyMax = (_n32DlyMax > 100000) ? 100000 : _n32DlyMax;
      }
      return(true);
    }
  }
  return(false);
}

} // namespace vge
