// ############################################################################################
//  CommonDialogPages.cc
//  Implementation of most commonly used dialog page classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../HighLevelClass/ApplicationObject.h"
#include "CommonDialogPages.h"


namespace vge {

//
// Message box
//
CMessageBox::CMessageBox(CVideoEnv          &iVideoEnv,
                         CEventHandlerBase  *eventHandler,
                         const string       &caption,
                         const string       &text,
                         EMsgBoxIcon         eIcon,
                         EMsgBoxButtons      eButton,
                         EMsgBoxSize         eSize)
: CDialogPage(iVideoEnv, NULL, eventHandler),
  _caption   (caption),
  _text      (text),
  _eIcon     (eIcon),
  _eButton   (eButton),
  _eSize     (eSize),
  _exit      (false)
{
  InitializeAllControls();
  RedrawAllControls();
  _iVideoEnv->UpdateDisplay();
}

CMessageBox::~CMessageBox() throw()
{
}

void CMessageBox::InitializeAllControls()
{
  const uint32_t ww = _iVideoEnv->WindowWidth();
  const uint32_t wh = _iVideoEnv->WindowHeight();

  // Calculate the size of the message box
  uint32_t pw, ph;
       if(_eSize == EMBS_Tiny)   { pw = 250; ph = 130; }
  else if(_eSize == EMBS_Small)  { pw = 440; ph = 220; }
  else if(_eSize == EMBS_Medium) { pw = 600; ph = 300; }
  else if(_eSize == EMBS_Large)  { pw = 760; ph = 380; }
  else if(_eSize == EMBS_Huge)   { pw = 984; ph = 492; }
  else                           { pw = ww;  ph = wh;  }
  if(pw > ww) pw = ww;
  if(ph > wh) ph = wh;

  const uint32_t px = (ww - pw) / 2;
  const uint32_t py = (wh - ph) / 2;

  // Initialize controls
  InitControl(_pbrIB, NULL, px, py, pw, ph, _caption.c_str());
  InitCompound(_hvgIB, &_pbrIB);
      _hvgIB.AddRowsEx(2, 0, 25);
      _hvgIB[0].AddBlocksEx(2, 33, 0);
  switch(_eButton) {
    case EMBB_OKCancel:
      _hvgIB[1].AddBlocks(2);
      InitControl(_btn1, _hvgIB[1], 0, _M("OK"), false, true);
          Connect(_btn1, CDP_SLOT(CMessageBox, _onBtnOK));
      InitControl(_btn2, _hvgIB[1], 1, _M("Cancel"), true, false);
          Connect(_btn2, CDP_SLOT(CMessageBox, _onBtnCancel));
      break;
    case EMBB_YesNo:
      _hvgIB[1].AddBlocks(2);
      InitControl(_btn1, _hvgIB[1], 0, _M("Yes"), false, true);
          Connect(_btn1, CDP_SLOT(CMessageBox, _onBtnYes));
      InitControl(_btn2, _hvgIB[1], 1, _M("No"), true, false);
          Connect(_btn2, CDP_SLOT(CMessageBox, _onBtnNo));
      break;
    case EMBB_YesNoCancel:
      _hvgIB[1].AddBlocks(3);
      InitControl(_btn1, _hvgIB[1], 0, _M("Yes"), false, true);
          Connect(_btn1, CDP_SLOT(CMessageBox, _onBtnYes));
      InitControl(_btn2, _hvgIB[1], 1, _M("No"), false, false);
          Connect(_btn2, CDP_SLOT(CMessageBox, _onBtnNo));
      InitControl(_btn3, _hvgIB[1], 2, _M("Cancel"), true, false);
          Connect(_btn3, CDP_SLOT(CMessageBox, _onBtnCancel));
      break;
    case EMBB_AbortRetry:
      _hvgIB[1].AddBlocks(2);
      InitControl(_btn1, _hvgIB[1], 0, _M("Abort"), true, false);
          Connect(_btn1, CDP_SLOT(CMessageBox, _onBtnAbort));
      InitControl(_btn2, _hvgIB[1], 1, _M("Retry"), false, true);
          Connect(_btn2, CDP_SLOT(CMessageBox, _onBtnRetry));
      break;
    case EMBB_AbortRetryIgnore:
      _hvgIB[1].AddBlocks(3);
      InitControl(_btn1, _hvgIB[1], 0, _M("Abort"), true, false);
          Connect(_btn1, CDP_SLOT(CMessageBox, _onBtnAbort));
      InitControl(_btn2, _hvgIB[1], 1, _M("Retry"), false, true);
          Connect(_btn2, CDP_SLOT(CMessageBox, _onBtnRetry));
      InitControl(_btn3, _hvgIB[1], 2, _M("Ignore"), false, false);
          Connect(_btn3, CDP_SLOT(CMessageBox, _onBtnIgnore));
      break;
    default: // EMBB_OK
      InitControl(_btn1, _hvgIB[1], CGLayout::WHOLE_AREA, _M("OK"), true, true);
          Connect(_btn1, CDP_SLOT(CMessageBox, _onBtnOK));
      break;
  }

  // Calculate the text offset
  _iVideoEnv->GetStyle(&_style);
#ifndef __USE_UTF8_DIALOG_STRING__
  uint32_t a, d;
  _iVideoEnv->SetFont(_style.ptFont);
  a = _iVideoEnv->GetFontAscent();
  d = _iVideoEnv->GetFontDescent();
  _textYOffset = (a > d) ? (a - d) : a;
#else
  int32_t  x, y;
  uint32_t w, h;
  _iVideoEnv->SetUtf8Font(_style.ptFontSet);
  _iVideoEnv->GetUtf8TextGeometry(x, y, w, h, "Gg");
  _textYOffset = (y > 0) ? y : -y;
#endif
}

void CMessageBox::HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str)
{
  UNUSED_PARAM(str);

  switch(_eButton) {
    case EMBB_OKCancel:
           if(keySymbol==VKS_ENTER || keySymbol==VKS_KP_ENTER) _onBtnOK(NULL);
      else if(keySymbol==VKS_ESC)                              _onBtnCancel(NULL);
      break;
    case EMBB_YesNo:
           if(keySymbol==VKS_ENTER || keySymbol==VKS_KP_ENTER) _onBtnYes(NULL);
      else if(keySymbol==VKS_ESC)                              _onBtnNo(NULL);
      break;
    case EMBB_YesNoCancel:
           if(keySymbol==VKS_ENTER || keySymbol==VKS_KP_ENTER) _onBtnYes(NULL);
      else if(keySymbol==VKS_ESC)                              _onBtnCancel(NULL);
      break;
    case EMBB_AbortRetry:
           if(keySymbol==VKS_ENTER || keySymbol==VKS_KP_ENTER) _onBtnRetry(NULL);
      else if(keySymbol==VKS_ESC)                              _onBtnAbort(NULL);
      break;
    case EMBB_AbortRetryIgnore:
           if(keySymbol==VKS_ENTER || keySymbol==VKS_KP_ENTER) _onBtnRetry(NULL);
      else if(keySymbol==VKS_ESC)                              _onBtnAbort(NULL);
      break;
    default: // EMBB_OK
      if(keySymbol==' '          ||
         keySymbol==VKS_ENTER    ||
         keySymbol==VKS_KP_ENTER ||
         keySymbol==VKS_ESC) _onBtnOK(NULL);
      break;
  }
}

void CMessageBox::_onBtnOK(_CControl *ctrl)
{
  UNUSED_PARAM(ctrl);
  _resultCode = EMBRV_OK;
  _exit = true; 
}

void CMessageBox::_onBtnCancel(_CControl *ctrl)
{
  UNUSED_PARAM(ctrl);
  _resultCode = EMBRV_Cancel;
  _exit = true; 
}

void CMessageBox::_onBtnYes(_CControl *ctrl)
{
  UNUSED_PARAM(ctrl);
  _resultCode = EMBRV_Yes;
  _exit = true; 
}

void CMessageBox::_onBtnNo(_CControl *ctrl)
{
  UNUSED_PARAM(ctrl);
  _resultCode = EMBRV_No;
  _exit = true; 
}

void CMessageBox::_onBtnAbort(_CControl *ctrl)
{
  UNUSED_PARAM(ctrl);
  _resultCode = EMBRV_Abort;
  _exit = true; 
}

void CMessageBox::_onBtnRetry(_CControl *ctrl)
{
  UNUSED_PARAM(ctrl);
  _resultCode = EMBRV_Retry;
  _exit = true; 
}

void CMessageBox::_onBtnIgnore(_CControl *ctrl)
{
  UNUSED_PARAM(ctrl);
  _resultCode = EMBRV_Ignore;
  _exit = true; 
}

void CMessageBox::CustomDraw_OnRedrawAllControls ()
{
  _iVideoEnv->DrawMsgBoxIcon(_pbrIB.GetXClient() + _hvgIB[0][0].x1 + CGLayout::DEFAULT_PAD_SIZE,
                             _pbrIB.GetYClient() + _hvgIB[0][0].y1 + CGLayout::DEFAULT_PAD_SIZE, _eIcon);

  _iVideoEnv->SetForegroundColor(_style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
  _iVideoEnv->SetFont(_style.ptFont);
  _iVideoEnv->DrawTextML(_pbrIB.GetXClient() + _hvgIB[0][1].x1 + CGLayout::DEFAULT_PAD_SIZE + 2,
                         _pbrIB.GetYClient() + _hvgIB[0][1].y1 + CGLayout::DEFAULT_PAD_SIZE + 2 + _textYOffset,
                         _text);
#else
  _iVideoEnv->SetUtf8Font(_style.ptFontSet);
  _iVideoEnv->DrawUtf8TextML(_pbrIB.GetXClient() + _hvgIB[0][1].x1 + CGLayout::DEFAULT_PAD_SIZE + 2,
                             _pbrIB.GetYClient() + _hvgIB[0][1].y1 + CGLayout::DEFAULT_PAD_SIZE + 2 + _textYOffset,
                             _text);
#endif
}

void CMessageBox::CustomDraw_OnButtonPress(EBtnSymbol buttonSymbol)
{
  UNUSED_PARAM(buttonSymbol);
  CustomDraw_OnRedrawAllControls();
}

void CMessageBox::CustomDraw_OnButtonRelease(EBtnSymbol buttonSymbol)
{
  UNUSED_PARAM(buttonSymbol);
  CustomDraw_OnRedrawAllControls();
}

void CMessageBox::CustomDraw_OnMouseMove(int32_t pointerXPos, int32_t pointerYPos)
{
  UNUSED_PARAMS_2(pointerXPos, pointerYPos);
  CustomDraw_OnRedrawAllControls();
}


//
// Single-line input box
//
CInputBoxSL::CInputBoxSL(CVideoEnv           &iVideoEnv,
                         CEventHandlerBase   *eventHandler,
                         const string        &caption,
                         const string        &defaultText,
                         EMsgBoxSize          eSize,
                         ETBType              sltbType,
                         bool                 sltbEscCancel,
                         PFNSLTextBoxCallback pfnSLTBUserCallback)
: CDialogPage   (iVideoEnv, NULL, eventHandler),
  _caption      (caption),
  _defaultText  (defaultText),
  _eSize        (eSize),
  _sltbType     (sltbType),
  _sltbEscCancel(sltbEscCancel),
  _pfnSLTBUC    (pfnSLTBUserCallback),
  _exit         (false)
{
  InitializeAllControls();
  RedrawAllControls();
  _iVideoEnv->UpdateDisplay();
}

CInputBoxSL::~CInputBoxSL() throw()
{
}

void CInputBoxSL::InitializeAllControls()
{
  const uint32_t ww = _iVideoEnv->WindowWidth();
  const uint32_t wh = _iVideoEnv->WindowHeight();

  // Calculate the size of the message box
  uint32_t pw, ph;
       if(_eSize == EMBS_Tiny)   { pw = 250; ph = 95; }
  else if(_eSize == EMBS_Small)  { pw = 440; ph = 95; }
  else if(_eSize == EMBS_Medium) { pw = 600; ph = 95; }
  else if(_eSize == EMBS_Large)  { pw = 760; ph = 95; }
  else if(_eSize == EMBS_Huge)   { pw = 984; ph = 95; }
  else                           { pw = ww;  ph = 95; }
  if(pw > ww) pw = ww;
  if(ph > wh) ph = wh;

  const uint32_t px = (ww - pw) / 2;
  const uint32_t py = (wh - ph) / 2;

  // Initialize controls
  InitControl(_pbrIB, NULL, px, py, pw, ph, _caption.c_str());
  InitCompound(_hvgIB, &_pbrIB);
      _hvgIB.AddRowsEx(2, 0, 25);
      _hvgIB[1].AddBlocks(2);
  InitControl(_txtIB, _hvgIB[0], CGLayout::WHOLE_AREA, _sltbType, _sltbEscCancel, _pfnSLTBUC);
      _txtIB.SetText(_defaultText.c_str());
  InitControl(_btnOK, _hvgIB[1], 0, _M("OK"), false, true);
      Connect(_btnOK, CDP_SLOT(CInputBoxSL, _onBtnOK));
  InitControl(_btnCancel, _hvgIB[1], 1, _M("Cancel"), true, false);
      Connect(_btnCancel, CDP_SLOT(CInputBoxSL, _onBtnCancel));
}

void CInputBoxSL::HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str)
{
  UNUSED_PARAM(str);
  if(keySymbol==VKS_ENTER || keySymbol==VKS_KP_ENTER)
    _onBtnOK(NULL);
  else if(keySymbol == VKS_ESC)
    _onBtnCancel(NULL);
}

void CInputBoxSL::_onBtnOK(_CControl *ctrl)
{
  UNUSED_PARAM(ctrl);
  _resultText = _txtIB.GetText(); 
  _exit = true; 
}

void CInputBoxSL::_onBtnCancel(_CControl *ctrl)
{
  UNUSED_PARAM(ctrl);
  _resultText = ""; 
  _exit = true; 
}


//
// Multi-line input box
//
CInputBoxML::CInputBoxML(CVideoEnv           &iVideoEnv,
                         CEventHandlerBase   *eventHandler,
                         const string        &caption,
                         const string        &defaultText,
                         EMsgBoxSize          eSize,
                         bool                 mltbEscCancel,
                         PFNMLTextBoxCallback pfnmltbUserCallback)
: CDialogPage   (iVideoEnv, NULL, eventHandler),
  _caption      (caption),
  _defaultText  (defaultText),
  _eSize        (eSize),
  _mltbEscCancel(mltbEscCancel),
  _pfnMLTBUC    (pfnmltbUserCallback),
  _exit         (false)
{
  InitializeAllControls();
  RedrawAllControls();
  _iVideoEnv->UpdateDisplay();
}

CInputBoxML::~CInputBoxML() throw()
{
}

void CInputBoxML::InitializeAllControls()
{
  const uint32_t ww = _iVideoEnv->WindowWidth();
  const uint32_t wh = _iVideoEnv->WindowHeight();

  // Calculate the size of the message box
  uint32_t pw, ph;
       if(_eSize == EMBS_Tiny)   { pw = 250; ph = 130; }
  else if(_eSize == EMBS_Small)  { pw = 440; ph = 220; }
  else if(_eSize == EMBS_Medium) { pw = 600; ph = 300; }
  else if(_eSize == EMBS_Large)  { pw = 760; ph = 380; }
  else if(_eSize == EMBS_Huge)   { pw = 984; ph = 492; }
  else                           { pw = ww;  ph = wh;  }
  if(pw > ww) pw = ww;
  if(ph > wh) ph = wh;

  const uint32_t px = (ww - pw) / 2;
  const uint32_t py = (wh - ph) / 2;

  // Initialize controls
  InitControl(_pbrIB, NULL, px, py, pw, ph, _caption.c_str());
  InitCompound(_hvgIB, &_pbrIB);
      _hvgIB.AddRowsEx(2, 0, 25);
      _hvgIB[1].AddBlocks(2);
  InitControl(_txtIB, _hvgIB[0], CGLayout::WHOLE_AREA, _mltbEscCancel, _pfnMLTBUC);
      _txtIB.SetText(_defaultText.c_str());
  InitControl(_btnOK, _hvgIB[1], 0, _M("OK"), false, true);
      Connect(_btnOK, CDP_SLOT(CInputBoxML, _onBtnOK));
  InitControl(_btnCancel, _hvgIB[1], 1, _M("Cancel"), true, false);
      Connect(_btnCancel, CDP_SLOT(CInputBoxML, _onBtnCancel));
}

void CInputBoxML::HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str)
{
  UNUSED_PARAM(str);
  if(keySymbol==VKS_ENTER || keySymbol==VKS_KP_ENTER)
    _onBtnOK(NULL);
  else if(keySymbol == VKS_ESC)
    _onBtnCancel(NULL);
}

void CInputBoxML::_onBtnOK(_CControl *ctrl)
{
  UNUSED_PARAM(ctrl);
  _resultText = _txtIB.GetText(); 
  _exit = true; 
}

void CInputBoxML::_onBtnCancel(_CControl *ctrl)
{
  UNUSED_PARAM(ctrl);
  _resultText = ""; 
  _exit = true; 
}

} // namespace vge
