// ############################################################################################
//  DialogLayout.cc
//  Implementation of VortexGE layout controls
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "DialogLayout.h"
#include "VideoEnvironment.h"


namespace vge {

//
// Generic-layout control
//
const uint32_t CGLayout::DEFAULT_PAD_SIZE;
const uint32_t CGLayout::WHOLE_AREA;


CGLayout::~CGLayout() throw()
{
}

void CGLayout::InitControl(CVideoEnv &vidEnv, const _CControl *parent, size_t parentBlockIndex, uint32_t borderWidth, uint32_t spacing)
{
  if(!parent)
    InitBaseData(vidEnv, parent, _CtrlID(), 0, 0, vidEnv.WindowWidth(), vidEnv.WindowHeight());
  else {
    const CPanel    *pn = dynamic_cast<const CPanel*>(parent);
    const CPanelBar *pb = dynamic_cast<const CPanelBar*>(parent);
    const CGLayout  *gl = dynamic_cast<const CGLayout*>(parent);
    if(pn)
      InitBaseData(vidEnv, pn, _CtrlID(), pn->GetXClient(false), pn->GetYClient(false), pn->GetWClient(), pn->GetHClient());
    else if(pb)
      InitBaseData(vidEnv, pb, _CtrlID(), pb->GetXClient(false), pb->GetYClient(false), pb->GetWClient(), pb->GetHClient());
    else if(gl) {
      const CGLayout &l = *gl;
      const size_t   &d = parentBlockIndex;
      InitBaseData(vidEnv, gl, _CtrlID(), l[d].x1, l[d].y1, l[d].w, l[d].h);
    }
    else
      VError(VERR_InvalidParameter, _NFL_, "Only the CPanel, CPanelBar or another layout class can be the parent of a layout class");
  }

  // Initialize variables
  _spacing     = spacing;
  _borderWidth = borderWidth;
  _marea.x1    = _borderWidth;
  _marea.y1    = _borderWidth;
  _marea.x2    = GetWidth() - _borderWidth;
  _marea.y2    = GetHeight() - _borderWidth;
  _marea.w     = _marea.x2 - _borderWidth;
  _marea.h     = _marea.y2 - _borderWidth;
}

void CGLayout::AddBlocks(size_t count, uint32_t size)
{
  if(!count) return;
  for(size_t i = 0; i < count; ++i) _blocks.push_back(size);
  _bdirty = true;
}

void CGLayout::AddBlocksEx(size_t count, uint32_t size, ...)
{
  if(!count) return;

  _blocks.push_back(size);

  va_list vl;
  va_start(vl, size);
    while(--count) _blocks.push_back(va_arg(vl, uint32_t));
  va_end(vl);

  _bdirty = true;
}

const CGLayout::TBlock &CGLayout::operator[](uint32_t index) const
{
  if(_blocks.empty() || index==WHOLE_AREA) return(_marea);
  if(index > _blocks.size()) VError(VERR_InvalidParameter, _NFL_, "Block index out of range");
  
  if(_bdirty) {
    _CalcBlocksAreas();
    _bdirty = false;
  }
  return(_bareas[index]);
}


//
// Horizontal-layout control
//
CHLayout::~CHLayout() throw()
{
}

void CHLayout::_CalcBlocksAreas() const
{
  _bareas.clear();
  if(_blocks.empty()) return;

  // Calculate the cumulative "fixed height"
  uint32_t hCum = 0;
  uint32_t hDiv = 0;
  for(size_t i = 0; i < _blocks.size(); ++i) {
    hCum += _blocks[i];
    if(!_blocks[i]) hDiv++;
  }
  if(!hDiv) hDiv = 1;

  // Calculate the "equally divided height"
  const uint32_t hRem = _marea.h - hCum - _spacing * (_blocks.size() - 1);
  const uint32_t hEql = hRem / hDiv;
        uint32_t hRes = hRem % hDiv;

  // Calculate the area
  int32_t yc = _marea.y1;
  for(size_t i = 0; i < _blocks.size(); ++i) {
    int32_t hc = _blocks[i];
    if(!hc) {
      hc = hEql;
      if(hRes) {
        hc++;
       hRes--;
      }
    }

    _bareas.push_back(TBlock(_marea.x1, yc, _marea.x2, yc + hc - 1, _marea.w, hc));

    yc += (hc + _spacing);
  }
}


//
// Vertical-layout control
//
CVLayout::~CVLayout() throw()
{
}

void CVLayout::_CalcBlocksAreas() const
{
  _bareas.clear();
  if(_blocks.empty()) return;

  // Calculate the cumulative "fixed width"
  uint32_t wCum = 0;
  uint32_t wDiv = 0;
  for(size_t i = 0; i < _blocks.size(); ++i) {
    wCum += _blocks[i];
    if(!_blocks[i]) wDiv++;
  }
  if(!wDiv) wDiv = 1;

  // Calculate the "equally divided width"
  const uint32_t wRem = _marea.w - wCum - _spacing * (_blocks.size() - 1);
  const uint32_t wEql = wRem / wDiv;
  const uint32_t wRes = wRem % wDiv;

  // Calculate the area
  int32_t xc = _marea.x1;
  for(size_t i = 0; i < _blocks.size(); ++i) {
    int32_t wc = _blocks[i] ? _blocks[i] : wEql;
    if(i == _blocks.size()-1) wc += wRes;

    _bareas.push_back(TBlock(xc, _marea.y1, xc + wc - 1, _marea.y2, wc, _marea.h));
    xc += (wc + _spacing);
  }
}


//
// HV-layout-compund (not a control)
//
CHVLayoutCompound::~CHVLayoutCompound() throw()
{
  _DeleteVLayouts();
}

void CHVLayoutCompound::_DeleteVLayouts()
{
  for(size_t i = 0; i < _vls.size(); ++i) delete _vls[i];
  _vls.clear();
}

void CHVLayoutCompound::InitCompound(CVideoEnv &vidEnv, const _CControl *parent, size_t parentBlockIndex, uint32_t borderWidth, uint32_t spacing)
{
  _vidEnv  = &vidEnv;
  _spacing = spacing;
  _hls.InitControl(vidEnv, parent, parentBlockIndex, borderWidth, spacing);
}

void CHVLayoutCompound::AddRow(uint32_t rowSize)
{
  _hls.AddBlock(rowSize);
  _DeleteVLayouts();
}

void CHVLayoutCompound::AddRows(size_t rowCount, uint32_t rowSize)
{
  _hls.AddBlocks(rowCount, rowSize);
  _DeleteVLayouts();
}

void CHVLayoutCompound::AddRowsEx(size_t rowCount, uint32_t rowSize, ...)
{
  if(!rowCount) return;

  _hls.AddBlock(rowSize);

  va_list vl;
  va_start(vl, rowSize);
    while(--rowCount) _hls.AddBlock(va_arg(vl, uint32_t));
  va_end(vl);

  _DeleteVLayouts();
}

CVLayout &CHVLayoutCompound::_GetVLayout(size_t rowIndex) const
{
  if(_vls.empty()) {
    try {
      _vls.resize(_hls.GetBlockCount());
      for(size_t i = 0; i < _vls.size(); ++i) {
        _vls[i] = _nothrow_new CVLayout;
        if(!_vls[i]) VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for new CVLayout instance");
        _vls[i]->InitControl(*_vidEnv, &_hls, i, 0, _spacing);
      }
    }
    catch(...) {
      for(size_t i = 0; i < _vls.size(); ++i) delete _vls[i];
      _vls.clear();
      throw;
    }
  }

  if(rowIndex >= _vls.size()) VError(VERR_InvalidParameter, _NFL_, "Row index out of range");
  return(*_vls[rowIndex]);
}


//
// VH-layout-compound (not a control)
//
CVHLayoutCompound::~CVHLayoutCompound() throw()
{
  _DeleteHLayouts();
}

void CVHLayoutCompound::_DeleteHLayouts()
{
  for(size_t i = 0; i < _hls.size(); ++i) delete _hls[i];
  _hls.clear();
}

void CVHLayoutCompound::InitCompound(CVideoEnv &vidEnv, const _CControl *parent, size_t parentBlockIndex, uint32_t borderWidth, uint32_t spacing)
{
  _vidEnv  = &vidEnv;
  _spacing = spacing;
  _vls.InitControl(vidEnv, parent, parentBlockIndex, borderWidth, spacing);
}

void CVHLayoutCompound::AddCol(uint32_t colSize)
{
  _vls.AddBlock(colSize);
  _DeleteHLayouts();
}

void CVHLayoutCompound::AddCols(size_t colCount, uint32_t colSize)
{
  _vls.AddBlocks(colCount, colSize);
  _DeleteHLayouts();
}

void CVHLayoutCompound::AddColsEx(size_t colCount, uint32_t colSize, ...)
{
  if(!colCount) return;

  _vls.AddBlock(colSize);

  va_list vl;
  va_start(vl, colSize);
    while(--colCount) _vls.AddBlock(va_arg(vl, uint32_t));
  va_end(vl);

  _DeleteHLayouts();
}

CHLayout &CVHLayoutCompound::_GetHLayout(size_t colIndex) const
{
  if(_hls.empty()) {
    try {
      _hls.resize(_vls.GetBlockCount());
      for(size_t i = 0; i < _hls.size(); ++i) {
        _hls[i] = _nothrow_new CHLayout;
        if(!_hls[i]) VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for new CHLayout instance");
        _hls[i]->InitControl(*_vidEnv, &_vls, i, 0, _spacing);
      }
    }
    catch(...) {
      for(size_t i = 0; i < _hls.size(); ++i) delete _hls[i];
     _hls.clear();
      throw;
    }
  }

  if(colIndex >= _hls.size()) VError(VERR_InvalidParameter, _NFL_, "Column index out of range");
  return(*_hls[colIndex]);
}

} // namespace vge
