// ############################################################################################
//  VX11Dialog2.cc
//  Implementation of dialog classes using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// NOTE: This file is meant to be included from within
//       "lib/VideoEnvironment/DialogEnvironment.cc"


namespace vge {

//
// Command-button control
//
CButton::CButton()
: _img(0), _imgMask(0)
{
}

CButton::~CButton() throw()
{
  if(_img)     XFreePixmap(_tBD.pcsVidEnv->_tX11.ptDisplay, _img);
  if(_imgMask) XFreePixmap(_tBD.pcsVidEnv->_tX11.ptDisplay, _imgMask);
}

void CButton::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                          int32_t n32X, int32_t n32Y,
                          uint32_t un32Width, uint32_t un32Height,
                          const char *pcText, bool bInsideOutline, bool bThickOutline)
{
  // Parse the string
  string ps;
  const int32_t up = pcText ? _ParseDBT_(ps, pcText) : -1;

  // Basic initializations
  int32_t n32X1, n32Y1, n32X2, n32Y2;
  InitBaseData(vidEnv, pcsParent, ECID_Button, n32X, n32Y, un32Width, un32Height);
  SetString(ps.c_str());
  n32X1 = _tBD.n32X1; n32X2 = _tBD.n32X2;
  n32Y1 = _tBD.n32Y1; n32Y2 = _tBD.n32Y2;

  // Calculate button's coordinates
  if(bThickOutline) {
    n32X1++; n32Y1++;
    n32X2--; n32Y2--;
    _bThickOutline = true;
  }
  else
    _bThickOutline = false;
  _CalcPanelCoords_(&_tOut, _atSegTL, _atSegBR, &_tFace, n32X1, n32Y1, n32X2, n32Y2);
  if(bInsideOutline) {
    _tIn.x = _tFace.x++;
    _tIn.y = _tFace.y++;
    _tFace.width--;
    _tFace.height--;
    _tIn.width  = _tFace.width--;
    _tIn.height = _tFace.height--;
    _bInsideOutline = true;
  }
  else
    _bInsideOutline = false;
  _tOutThick.x = _tOut.x - 1;
  _tOutThick.y = _tOut.y - 1;
  _tOutThick.width  = _tOut.width + 2;
  _tOutThick.height = _tOut.height + 2;

  // Calculate button's text properties
#ifndef __USE_UTF8_DIALOG_STRING__
  _n32Xt = n32X1 + (n32X2 - n32X1 - XTextWidth(*vidEnv._tX11Style.ptFont, pcText, StringLen())) / 2;
  _n32Yt = n32Y1 + (n32Y2 - n32Y1 - (*vidEnv._tX11Style.ptFont)->ascent - (*vidEnv._tX11Style.ptFont)->descent) / 2 + (*vidEnv._tX11Style.ptFont)->ascent;
#else
  XRectangle tl;
  Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, pcText, StringLen(), NULL, &tl);
  _n32Xt = n32X1 + (n32X2 - n32X1 - tl.width) / 2;
  _n32Yt = n32Y1 + (n32Y2 - n32Y1 - tl.height) / 2 - tl.y;
#endif

  // Calculate the caption's underline position & width
  if(up >= 0) {
#ifndef __USE_UTF8_DIALOG_STRING__
    _ulX1 = _n32Xt - 1 + XTextWidth(*vidEnv._tX11Style.ptFont, ps.data(), up);
    _ulX2 = _n32Xt - 1 + XTextWidth(*vidEnv._tX11Style.ptFont, ps.data(), up + 1);
    _ulY  = _n32Yt + 1;
#else
    Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, ps.data(), up, NULL, &tl);
    _ulX1 = _n32Xt - 1 + tl.width;
    Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, ps.data(), up + 1, NULL, &tl);
    _ulX2 = _n32Xt - 1 + tl.width;
    _ulY  = _n32Yt + 1;
#endif
  }
  else
    _ulY = -1;
}

void CButton::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                          int32_t n32X, int32_t n32Y,
                          uint32_t un32Width, uint32_t un32Height,
                          const char *xpmDataBuffer[], bool bInsideOutline, bool bThickOutline)
{
  InitControl(vidEnv, pcsParent, n32X, n32Y, un32Width, un32Height, "", bInsideOutline, bThickOutline);

  uint32_t w, h;
  vidEnv._LoadXPMFromMemory_(xpmDataBuffer, &_img, &_imgMask, &w, &h);
  if(w>(un32Width-2) || h>(un32Height-2)) 
    VError(VERR_InvalidParameter, _NFL_, "The given XPM image must not larger than (control's_width-2)x(control's_height-2) pixels.");

  _n32Xt = _tBD.n32X1 + (un32Width - w) / 2;
  _n32Yt = _tBD.n32Y1 + (un32Height - h) / 2;
}

void CButton::DrawControl() const
{
  DrawControl(false);
}

void CButton::DrawControl(bool bPressed) const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw button's face
  vidEnv.SetFillStyle(FILL_SOLID);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
  if(vidEnv._tX11Style.xFillStyle != FILL_GRADIENT) {
    XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  _tFace.x, _tFace.y, _tFace.width, _tFace.height);
  }
  else {
    vidEnv.DrawGradientBox(_tFace.x, _tFace.y, _tFace.width, _tFace.height,
                           vidEnv._tX11Style.xLightBorder,
                           vidEnv._tX11Style.xDarkBorder,
                           vidEnv._tX11Style.xFace);
  }

  // Draw button's border
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xOutline);
  XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
    _tOut.x, _tOut.y, _tOut.width, _tOut.height);
  if(_bThickOutline) {
    XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
      _tOutThick.x, _tOutThick.y, _tOutThick.width, _tOutThick.height);
  }
  if(_bInsideOutline) {
    XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
      _tIn.x, _tIn.y, _tIn.width, _tIn.height);
  }
  if(bPressed && _tBD.bEnabled) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
    XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  const_cast<XSegment*>(_atSegTL), 4);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  const_cast<XSegment*>(_atSegBR), 4);
  }
  else {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  const_cast<XSegment*>(_atSegTL), 4);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
    XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  const_cast<XSegment*>(_atSegBR), 4);
  }

  // Draw button's image
  if(_img && _imgMask) {
    XSetClipOrigin(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC, _n32Xt, _n32Yt);
    XSetClipMask(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC, _imgMask);
    XCopyArea(vidEnv._tX11.ptDisplay, _img, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
              0, 0, 16, 16, _n32Xt, _n32Yt);
    XSetClipMask(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC, 0);
    if(!_tBD.bEnabled) {
      vidEnv.SetFillStyle(FILL_STIPPLED);
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
      XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                     _tFace.x, _tFace.y, _tFace.width, _tFace.height);
      vidEnv.SetFillStyle(FILL_SOLID);
    }
    // We won't draw the button's text
    return;
  }

  // Draw button's text
  if(_tBD.bEnabled) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
    vidEnv.SetFont(vidEnv._tX11Style.ptFont);
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                 _n32Xt, _n32Yt, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt, _n32Yt, GetString(), StringLen());
#endif
    if(_ulY >= 0) {
      XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, _ulX1, _ulY, _ulX2, _ulY);
    }
  }
  else {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt, _n32Yt, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt, _n32Yt, GetString(), StringLen());
#endif
    if(_ulY >= 0) {
      XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, _ulX1, _ulY, _ulX2, _ulY);
    }
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt - 1, _n32Yt - 1, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt - 1, _n32Yt - 1, GetString(), StringLen());
#endif
    if(_ulY >= 0) {
      XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, _ulX1 - 1, _ulY - 1, _ulX2 - 1, _ulY - 1);
    }
  }
}

void CButton::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  if(MouseOver(n32MouseX, n32MouseY))
    DrawControl(bPressed);
  else
    DrawControl(false);
}


//
// Check-box control
//
CCheckBox::~CCheckBox() throw()
{
}

void CCheckBox::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                            int32_t n32X, int32_t n32Y, const char *pcCaption,
                            bool bOriginatedFromRight)
{
  // Parse the string
  string ps;
  const int32_t up = pcCaption ? _ParseDBT_(ps, pcCaption) : -1;

  // Calculate size
  uint32_t un32W, un32H, un32F;
#ifndef __USE_UTF8_DIALOG_STRING__
  un32W = XTextWidth(*vidEnv._tX11Style.ptFont, ps.data(), ps.length()) + 11 + 5;
  un32H = (*vidEnv._tX11Style.ptFont)->ascent + (*vidEnv._tX11Style.ptFont)->descent;
  un32F = ((*vidEnv._tX11Style.ptFont)->ascent - (*vidEnv._tX11Style.ptFont)->descent) / 2;
#else
  const char *ct = ps.c_str();
  if(ct[0] == 0) ct = _DTM;
  XRectangle tl;
  Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, ct, strlen(ct), NULL, &tl);
  un32W = tl.width + 11 + 5;
  un32H = tl.height;
  un32F = -tl.y / 2; //labs(tl.height + tl.y) / 2;
#endif
  if(un32H < 11) un32H = 11;

  // Basic initializations
  if(bOriginatedFromRight)
    InitBaseData(vidEnv, pcsParent, ECID_CheckBox, n32X - un32W + 1, n32Y, un32W, un32H);
  else
    InitBaseData(vidEnv, pcsParent, ECID_CheckBox, n32X, n32Y, un32W, un32H);
  SetString(ps.c_str());
  _bChecked = false;

  // Calculate coordinates
  n32X = _tBD.n32X1;
  n32Y = _tBD.n32Y1;
  if(bOriginatedFromRight) {
    _n32Xt        = n32X;
    _n32Yt        = n32Y + (un32H / 2) + un32F;
    _tOut1.x      = n32X + un32W - 11; _tOut1.y      = n32Y + (un32H / 2) - 5;
    _tOut1.width  = 11;                _tOut1.height = 11;
    _tOut2.x      = _tOut1.x + 1;      _tOut2.y      = _tOut1.y + 1;
    _tOut2.width  = _tOut1.width - 2;  _tOut2.height = _tOut1.height - 2;
    _tBox1.x      = _tOut2.x + 1;      _tBox1.y      = _tOut2.y + 1;
    _tBox1.width  = _tOut2.width - 2;  _tBox1.height = _tOut2.height - 2;
    _tBox2.x      = _tBox1.x + 1;      _tBox2.y      = _tBox1.y + 1;
    _tBox2.width  = _tBox1.width - 2;  _tBox2.height = _tBox1.height - 2;
  }
  else {
    _tOut1.x      = n32X;             _tOut1.y      = n32Y + (un32H / 2) - 5;
    _tOut1.width  = 11;               _tOut1.height = 11;
    _tOut2.x      = _tOut1.x + 1;     _tOut2.y      = _tOut1.y + 1;
    _tOut2.width  = _tOut1.width - 2; _tOut2.height = _tOut1.height - 2;
    _tBox1.x      = _tOut2.x + 1;     _tBox1.y      = _tOut2.y + 1;
    _tBox1.width  = _tOut2.width - 2; _tBox1.height = _tOut2.height - 2;
    _tBox2.x      = _tBox1.x + 1;     _tBox2.y      = _tBox1.y + 1;
    _tBox2.width  = _tBox1.width - 2; _tBox2.height = _tBox1.height - 2;
    _n32Xt        = n32X + 11 + 5;
    _n32Yt        = n32Y + (un32H / 2) + un32F;
  }

  // Calculate the caption's underline position & width
  if(up >= 0) {
#ifndef __USE_UTF8_DIALOG_STRING__
    _ulX1 = _n32Xt - 1 + XTextWidth(*vidEnv._tX11Style.ptFont, ps.data(), up);
    _ulX2 = _n32Xt - 1 + XTextWidth(*vidEnv._tX11Style.ptFont, ps.data(), up + 1);
    _ulY  = _n32Yt + 1;
#else
    Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, ps.data(), up, NULL, &tl);
    _ulX1 = _n32Xt - 1 + tl.width;
    Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, ps.data(), up + 1, NULL, &tl);
    _ulX2 = _n32Xt - 1 + tl.width;
    _ulY  = _n32Yt + 1;
#endif
  }
  else
    _ulY = -1;
}

bool CCheckBox::ToggleValue(int32_t n32MouseX, int32_t n32MouseY)
{
  if(MouseOver(n32MouseX, n32MouseY) && _tBD.bEnabled) {
    _bChecked = !_bChecked;
    return(true);
  }
  return(false);
}

void CCheckBox::DrawControl() const
{
  DrawControl(false);
}

void CCheckBox::DrawControl(bool bPressed) const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw box
  vidEnv.SetFillStyle(FILL_SOLID);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
  XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
    _tOut1.x, _tOut1.y, _tOut1.width, _tOut1.height);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
  XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
    _tOut2.x, _tOut2.y, _tOut2.width, _tOut2.height);
  if(bPressed && _tBD.bEnabled) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
    XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
      _tBox1.x, _tBox1.y, _tBox1.width, _tBox1.height);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xHotPress);
    XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
      _tBox2.x, _tBox2.y, _tBox2.width  + 1, _tBox2.height + 1);
  }
  else {
    if(_bChecked) {
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
      XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
        _tBox1.x, _tBox1.y, _tBox1.width, _tBox1.height);
      if(_tBD.bEnabled)
        vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
      else
        vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
      XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
        _tBox2.x, _tBox2.y, _tBox2.width + 1, _tBox2.height + 1);
    }
    else {
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
      XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
        _tBox1.x, _tBox1.y, _tBox1.width + 1, _tBox1.height + 1);
    }
  }

  // Draw caption
  if(_tBD.bEnabled) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
    vidEnv.SetFont(vidEnv._tX11Style.ptFont);
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt, _n32Yt, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt, _n32Yt, GetString(), StringLen());
#endif
    if(_ulY >= 0) {
      XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, _ulX1, _ulY, _ulX2, _ulY);
    }
  }
  else {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt, _n32Yt, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt, _n32Yt, GetString(), StringLen());
#endif
    if(_ulY >= 0) {
      XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, _ulX1, _ulY, _ulX2, _ulY);
    }
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt - 1, _n32Yt - 1, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt - 1, _n32Yt - 1, GetString(), StringLen());
#endif
    if(_ulY >= 0) {
      XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, _ulX1 - 1, _ulY - 1, _ulX2 - 1, _ulY - 1);
    }
  }
}

void CCheckBox::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  if(MouseOver(n32MouseX, n32MouseY))
    DrawControl(bPressed);
  else
    DrawControl(false);
}


//
// Radio-button control
//
CRadioButton::~CRadioButton() throw()
{
}

void CRadioButton::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                               int32_t n32X, int32_t n32Y, const char *pcCaption,
                               bool bOriginatedFromRight)
{
  // Parse the string
  string ps;
  const int32_t up = pcCaption ? _ParseDBT_(ps, pcCaption) : -1;

  // Calculate size
  uint32_t un32W, un32H, un32F;
#ifndef __USE_UTF8_DIALOG_STRING__
  un32W = XTextWidth(*vidEnv._tX11Style.ptFont, ps.data(), ps.length()) + 11 + 5;
  un32H = (*vidEnv._tX11Style.ptFont)->ascent + (*vidEnv._tX11Style.ptFont)->descent;
  un32F = ((*vidEnv._tX11Style.ptFont)->ascent - (*vidEnv._tX11Style.ptFont)->descent) / 2;
#else
  const char *ct = ps.c_str();
  if(ct[0] == 0) ct = _DTM;
  XRectangle tl;
  Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, ct, strlen(ct), NULL, &tl);
  un32W = tl.width + 11 + 5;
  un32H = tl.height;
  un32F = -tl.y / 2; //labs(tl.height + tl.y) / 2;
#endif
  if(un32H < 11) un32H = 11;

  // Basic initializations
  if(bOriginatedFromRight)
    InitBaseData(vidEnv, pcsParent, ECID_RadioButton, n32X - un32W + 1, n32Y, un32W, un32H);
  else
    InitBaseData(vidEnv, pcsParent, ECID_RadioButton, n32X, n32Y, un32W, un32H);
  SetString(ps.c_str());
  _bChecked = false;

  // Calculate coordinates
  n32X = _tBD.n32X1;
  n32Y = _tBD.n32Y1;
  if(bOriginatedFromRight) {
    _n32Xt        = n32X;
    _n32Yt        = n32Y + (un32H / 2) + un32F;
    _tOut1.x      = n32X + un32W - 11; _tOut1.y      = n32Y + (un32H / 2) - 5;
    _tOut1.width  = 11;                _tOut1.height = 11;
    _tOut2.x      = _tOut1.x + 1;      _tOut2.y      = _tOut1.y + 1;
    _tOut2.width  = _tOut1.width - 2;  _tOut2.height = _tOut1.height - 2;
    _tBox1.x      = _tOut2.x + 1;      _tBox1.y      = _tOut2.y + 1;
    _tBox1.width  = _tOut2.width - 2;  _tBox1.height = _tOut2.height - 2;
    _tBox2.x      = _tBox1.x + 2;      _tBox2.y      = _tBox1.y + 2;
    _tBox2.width  = _tBox1.width - 4;  _tBox2.height = _tBox1.height - 4;
  }
  else {
    _tOut1.x      = n32X;             _tOut1.y      = n32Y + (un32H / 2) - 5;
    _tOut1.width  = 11;               _tOut1.height = 11;
    _tOut2.x      = _tOut1.x + 1;     _tOut2.y      = _tOut1.y + 1;
    _tOut2.width  = _tOut1.width - 2; _tOut2.height = _tOut1.height - 2;
    _tBox1.x      = _tOut2.x + 1;     _tBox1.y      = _tOut2.y + 1;
    _tBox1.width  = _tOut2.width - 2; _tBox1.height = _tOut2.height - 2;
    _tBox2.x      = _tBox1.x + 2;     _tBox2.y      = _tBox1.y + 2;
    _tBox2.width  = _tBox1.width - 4; _tBox2.height = _tBox1.height - 4;
    _n32Xt        = n32X + 11 + 5;
    _n32Yt        = n32Y + (un32H / 2) + un32F;
  }

  // Calculate the caption's underline position & width
  if(up >= 0) {
#ifndef __USE_UTF8_DIALOG_STRING__
    _ulX1 = _n32Xt - 1 + XTextWidth(*vidEnv._tX11Style.ptFont, ps.data(), up);
    _ulX2 = _n32Xt - 1 + XTextWidth(*vidEnv._tX11Style.ptFont, ps.data(), up + 1);
    _ulY  = _n32Yt + 1;
#else
    Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, ps.data(), up, NULL, &tl);
    _ulX1 = _n32Xt - 1 + tl.width;
    Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, ps.data(), up + 1, NULL, &tl);
    _ulX2 = _n32Xt - 1 + tl.width;
    _ulY  = _n32Yt + 1;
#endif
  }
  else
    _ulY = -1;
}

bool CRadioButton::ToggleValue(int32_t n32MouseX, int32_t n32MouseY, vector<CRadioButton*> &csRBInGroup)
{
  if(MouseOver(n32MouseX, n32MouseY)) {
    // Check
    if(!_tBD.bEnabled) return(false);
    if(_bChecked) return(true);
    // Make myself checked and make all my friends unchecked
    _bChecked = true;
    for(size_t i = 0; i < csRBInGroup.size(); ++i) {
      if(csRBInGroup[i] != this) csRBInGroup[i]->_bChecked = false;
    }
    return(true);
  }
  return(false);
}

bool CRadioButton::ToggleValue(int32_t n32MouseX, int32_t n32MouseY, CRadioButton *pcsRBInGroup, ...)
{
  //! \todo This function is really mangled, needs to be fixed later ;)

  // Temporary variables
  CRadioButton *pcsTmp, *pcsC, *pcsO;
  va_list      vlTmp;

  // Check if the event is on my self
  if(MouseOver(n32MouseX, n32MouseY)) {
    // Check
    if(!_tBD.bEnabled) return(false);
    if(_bChecked) return(true);
    // Make myself checked and make all my friends unchecked
    _bChecked = true;
    if(pcsRBInGroup) {
      pcsRBInGroup->_bChecked = false;
      va_start(vlTmp, pcsRBInGroup);
      for(;;) {
        pcsTmp = va_arg(vlTmp, CRadioButton *);
        if(pcsTmp)
          pcsTmp->_bChecked = false;
        else
          break;
      }
      va_end(vlTmp);
    }
    return(true);
  }

  // The event may be is on may friend, so move trough all my friends
  if(pcsRBInGroup) {
    // Check if the event is on my 1st friend
    if(pcsRBInGroup->MouseOver(n32MouseX, n32MouseY)) {
      // Check
      if(!pcsRBInGroup->_tBD.bEnabled) return(false);
      if(pcsRBInGroup->_bChecked) return(true);
      // Make my 1st friend checked and make all my other friends unchecked
      pcsRBInGroup->_bChecked = true;
      va_start(vlTmp, pcsRBInGroup);
      for(;;) {
        pcsTmp = va_arg(vlTmp, CRadioButton *);
        if(pcsTmp)
          pcsTmp->_bChecked = false;
        else
          break;
      }
      va_end(vlTmp);
      // Make myself unchecked
      _bChecked = false;
      return(true);
    }
    // Not on my 1st friend, check my other friends
    else {
      // Move trough all my friends, except the 1st one
      pcsC = pcsO = NULL;
      va_start(vlTmp, pcsRBInGroup);
      for(;;) {
        pcsTmp = va_arg(vlTmp, CRadioButton *);
        if(pcsTmp) {
          // Check for the event
          if(pcsTmp->MouseOver(n32MouseX, n32MouseY)) {
            // Check
            if(!pcsTmp->_tBD.bEnabled) continue;
            if(pcsTmp->_bChecked) return(true);
            // Make my friend checked
            pcsTmp->_bChecked = true;
            // Save my friend address
            pcsC = pcsTmp;
          }
          else {
            // Check if it is previously checked
            if(pcsTmp->_bChecked) pcsO = pcsTmp;
            // Make my friend unchecked
            pcsTmp->_bChecked = false;
          }
        }
        else
          break;
      }
      va_end(vlTmp);
      // Check if any of my friend is just newly got checked
      if(pcsC) {
        // Make my 1st friend unchecked
        pcsRBInGroup->_bChecked = false;
        // Make myself unchecked
        _bChecked = false;
        return(true);
      }
      else {
        // Restore the checked state (if any)
        if(pcsO) pcsO->_bChecked = true;
      }
    }
  }
  return(false);
}

void CRadioButton::DrawControl() const
{
  DrawControl(false);
}

void CRadioButton::DrawControl(bool bPressed) const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw box
  vidEnv.SetFillStyle(FILL_SOLID);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
  XDrawArc(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
    _tOut1.x, _tOut1.y, _tOut1.width, _tOut1.height, 0, 360 * 64);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
  XDrawArc(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
    _tOut2.x, _tOut2.y, _tOut2.width, _tOut2.height, 0, 360 * 64);
  if(bPressed && _tBD.bEnabled) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
    XDrawArc(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
      _tBox1.x, _tBox1.y, _tBox1.width, _tBox1.height, 0, 360 * 64);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xHotPress);
    XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
      _tBox2.x, _tBox2.y, _tBox2.width + 1, _tBox2.height + 1);
  }
  else {
    if(_bChecked) {
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
      XDrawArc(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
        _tBox1.x, _tBox1.y, _tBox1.width, _tBox1.height, 0, 360 * 64);
      if(_tBD.bEnabled)
        vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
      else
        vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
      XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
        _tBox2.x, _tBox2.y, _tBox2.width + 1, _tBox2.height + 1);
    }
    else {
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
      XFillArc(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
        _tBox1.x, _tBox1.y, _tBox1.width, _tBox1.height, 0, 360 * 64);
    }
  }

  // Draw caption
  if(_tBD.bEnabled) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
    vidEnv.SetFont(vidEnv._tX11Style.ptFont);
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt, _n32Yt, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt, _n32Yt, GetString(), StringLen());
#endif
    if(_ulY >= 0) {
      XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, _ulX1, _ulY, _ulX2, _ulY);
    }
  }
  else {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt, _n32Yt, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt, _n32Yt, GetString(), StringLen());
#endif
    if(_ulY >= 0) {
      XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, _ulX1, _ulY, _ulX2, _ulY);
    }
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt - 1, _n32Yt - 1, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt - 1, _n32Yt - 1, GetString(), StringLen());
#endif
    if(_ulY >= 0) {
      XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, _ulX1 - 1, _ulY - 1, _ulX2 - 1, _ulY - 1);
    }
  }
}

void CRadioButton::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  if(MouseOver(n32MouseX, n32MouseY))
    DrawControl(bPressed);
  else
    DrawControl(false);
}


//
// Spin-button control
//
CSpinButton::~CSpinButton() throw()
{
}

void CSpinButton::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                              ESBStyle eSpinButtonType, EHAlignment eTextAligment,
                              ESBValue eValueType, bool bWrapValue)
{
  // Temporary variables
  int32_t n32X1, n32Y1, n32X2, n32Y2, n32Xc, n32Yc;

  // Check
  if(eSpinButtonType==ESBS_VLeft || eSpinButtonType==ESBS_VRight  ||
     eSpinButtonType==ESBS_HTop  || eSpinButtonType==ESBS_HBottom) {
    if(un32Width<28 || un32Height<28)
      VError(VERR_InvalidParameter, _NFL_,
             "Spin button size is too small\nThe minimum size for ESBS_VLeft, ESBS_VRight, ESBS_HTop and ESBS_HBottom are 28x28 pixels");
  }
  else {
    if(un32Width<41 || un32Height<15)
      VError(VERR_InvalidParameter, _NFL_,
             "Spin button size is too small\nThe minimum size for EVSBHLeft, EVSBHRight and EVSBHCenter are 41x15 pixels");
  }

  // Basic initializations
  InitBaseData(vidEnv, pcsParent, ECID_SpinButton, n32X, n32Y, un32Width, un32Height);
  n32X   = _tBD.n32X1;
  n32Y   = _tBD.n32Y1;
  _eVT   = eValueType;
  _eTA   = eTextAligment;
  _bWrap = bWrapValue;

  // Calculate coordinates
  _tOutline.x = n32X;
  _tOutline.y = n32Y;
  _tOutline.width  = un32Width - 1;
  _tOutline.height = un32Height - 1;
#ifndef __USE_UTF8_DIALOG_STRING__
  _n32Yt = n32Y + (un32Height / 2) + (((*vidEnv._tX11Style.ptFont)->ascent - (*vidEnv._tX11Style.ptFont)->descent) / 2);
#else
  XRectangle tl;
  Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, "0123456789+-. Ee", 16, NULL, &tl);
  _n32Yt = n32Y + (un32Height / 2) - (tl.y / 2);
#endif
  if(eSpinButtonType == ESBS_VLeft) {
    // Text box
    n32X1 = n32X + 1; n32X2 = n32X1 + un32Width - 13 - 2 - 1;
    n32Y1 = n32Y + 1; n32Y2 = n32Y1 + un32Height - 2 - 1;
    _atText[0].x1 = n32X1;     _atText[0].y1 = n32Y1;     _atText[0].x2 = n32X1; _atText[0].y2 = n32Y2;
    _atText[1].x1 = n32X1 + 1; _atText[1].y1 = n32Y1;     _atText[1].x2 = n32X2; _atText[1].y2 = n32Y1;
    _atText[2].x1 = n32X1 + 1; _atText[2].y1 = n32Y2;     _atText[2].x2 = n32X2; _atText[2].y2 = n32Y2;
    _atText[3].x1 = n32X2;     _atText[3].y1 = n32Y1 + 1; _atText[3].x2 = n32X2; _atText[3].y2 = n32Y2 - 1;
    // Up button
    n32X1 = n32X + un32Width - 13 - 1; n32X2 = n32X1 + 13 - 1;
    n32Y1 = n32Y + 1;                  n32Y2 = n32Y1 + (un32Height / 2) - 2;
    _tUp.x1     = n32X1;     _tUp.y1     = n32Y1;     _tUp.x2     = n32X2; _tUp.y2     = n32Y2;
    _atUp[0].x1 = n32X1;     _atUp[0].y1 = n32Y1;     _atUp[0].x2 = n32X1; _atUp[0].y2 = n32Y2;
    _atUp[1].x1 = n32X1 + 1; _atUp[1].y1 = n32Y1;     _atUp[1].x2 = n32X2; _atUp[1].y2 = n32Y1;
    _atUp[2].x1 = n32X1 + 1; _atUp[2].y1 = n32Y2;     _atUp[2].x2 = n32X2; _atUp[2].y2 = n32Y2;
    _atUp[3].x1 = n32X2;     _atUp[3].y1 = n32Y1 + 1; _atUp[3].x2 = n32X2; _atUp[3].y2 = n32Y2 - 1;
    n32Xc = n32X1 + 6; n32Yc = n32Y1 + (un32Height / 4) - 1;
    _atArrow[0].x1 = n32Xc - 4; _atArrow[0].x2 = n32Xc;     _atArrow[0].y1 = n32Yc + 2; _atArrow[0].y2 = n32Yc - 2;
    _atArrow[1].x1 = n32Xc;     _atArrow[1].x2 = n32Xc + 4; _atArrow[1].y1 = n32Yc - 2; _atArrow[1].y2 = n32Yc + 2;
    _atArrow[4].x1 = _atArrow[0].x1; _atArrow[4].y1 = _atArrow[0].y1 + 1;
    _atArrow[4].x2 = _atArrow[0].x2; _atArrow[4].y2 = _atArrow[0].y2 + 1;
    _atArrow[5].x1 = _atArrow[1].x1; _atArrow[5].y1 = _atArrow[1].y1 + 1;
    _atArrow[5].x2 = _atArrow[1].x2; _atArrow[5].y2 = _atArrow[1].y2 + 1;
    // Down button
    n32X1 = n32X + un32Width - 13 - 1; n32X2 = n32X1 + 13 - 1;
    n32Y2 = n32Y + un32Height - 1 - 1; n32Y1 = n32Y2 - (un32Height / 2) + 2;
    _tDn.x1     = n32X1;     _tDn.y1     = n32Y1;     _tDn.x2     = n32X2; _tDn.y2     = n32Y2;
    _atDn[0].x1 = n32X1;     _atDn[0].y1 = n32Y1;     _atDn[0].x2 = n32X1; _atDn[0].y2 = n32Y2;
    _atDn[1].x1 = n32X1 + 1; _atDn[1].y1 = n32Y1;     _atDn[1].x2 = n32X2; _atDn[1].y2 = n32Y1;
    _atDn[2].x1 = n32X1 + 1; _atDn[2].y1 = n32Y2;     _atDn[2].x2 = n32X2; _atDn[2].y2 = n32Y2;
    _atDn[3].x1 = n32X2;     _atDn[3].y1 = n32Y1 + 1; _atDn[3].x2 = n32X2; _atDn[3].y2 = n32Y2 - 1;
    n32Xc = n32X1 + 6; n32Yc = n32Y1 + (un32Height / 4) - 1;
    _atArrow[2].x1 = n32Xc - 4; _atArrow[2].x2 = n32Xc;     _atArrow[2].y1 = n32Yc - 2; _atArrow[2].y2 = n32Yc + 2;
    _atArrow[3].x1 = n32Xc;     _atArrow[3].x2 = n32Xc + 4; _atArrow[3].y1 = n32Yc + 2; _atArrow[3].y2 = n32Yc - 2;
    _atArrow[6].x1 = _atArrow[2].x1; _atArrow[6].y1 = _atArrow[2].y1 + 1;
    _atArrow[6].x2 = _atArrow[2].x2; _atArrow[6].y2 = _atArrow[2].y2 + 1;
    _atArrow[7].x1 = _atArrow[3].x1; _atArrow[7].y1 = _atArrow[3].y1 + 1;
    _atArrow[7].x2 = _atArrow[3].x2; _atArrow[7].y2 = _atArrow[3].y2 + 1;
  }
  else if(eSpinButtonType == ESBS_VRight) {
    // Text box
    n32X1 = n32X + 1 + 13; n32X2 = n32X1 + un32Width - 13 - 2 - 1;
    n32Y1 = n32Y + 1;      n32Y2 = n32Y1 + un32Height - 2 - 1;
    _atText[0].x1 = n32X1;     _atText[0].y1 = n32Y1;     _atText[0].x2 = n32X1; _atText[0].y2 = n32Y2;
    _atText[1].x1 = n32X1 + 1; _atText[1].y1 = n32Y1;     _atText[1].x2 = n32X2; _atText[1].y2 = n32Y1;
    _atText[2].x1 = n32X1 + 1; _atText[2].y1 = n32Y2;     _atText[2].x2 = n32X2; _atText[2].y2 = n32Y2;
    _atText[3].x1 = n32X2;     _atText[3].y1 = n32Y1 + 1; _atText[3].x2 = n32X2; _atText[3].y2 = n32Y2 - 1;
    // Up button
    n32X1 = n32X + 1; n32X2 = n32X1 + 13 - 1;
    n32Y1 = n32Y + 1; n32Y2 = n32Y1 + (un32Height / 2) - 2;
    _tUp.x1     = n32X1;     _tUp.y1     = n32Y1;     _tUp.x2     = n32X2; _tUp.y2     = n32Y2;
    _atUp[0].x1 = n32X1;     _atUp[0].y1 = n32Y1;     _atUp[0].x2 = n32X1; _atUp[0].y2 = n32Y2;
    _atUp[1].x1 = n32X1 + 1; _atUp[1].y1 = n32Y1;     _atUp[1].x2 = n32X2; _atUp[1].y2 = n32Y1;
    _atUp[2].x1 = n32X1 + 1; _atUp[2].y1 = n32Y2;     _atUp[2].x2 = n32X2; _atUp[2].y2 = n32Y2;
    _atUp[3].x1 = n32X2;     _atUp[3].y1 = n32Y1 + 1; _atUp[3].x2 = n32X2; _atUp[3].y2 = n32Y2 - 1;
    n32Xc = n32X1 + 6; n32Yc = n32Y1 + (un32Height / 4) - 1;
    _atArrow[0].x1 = n32Xc - 4; _atArrow[0].x2 = n32Xc;     _atArrow[0].y1 = n32Yc + 2; _atArrow[0].y2 = n32Yc - 2;
    _atArrow[1].x1 = n32Xc;     _atArrow[1].x2 = n32Xc + 4; _atArrow[1].y1 = n32Yc - 2; _atArrow[1].y2 = n32Yc + 2;
    _atArrow[4].x1 = _atArrow[0].x1; _atArrow[4].y1 = _atArrow[0].y1 + 1;
    _atArrow[4].x2 = _atArrow[0].x2; _atArrow[4].y2 = _atArrow[0].y2 + 1;
    _atArrow[5].x1 = _atArrow[1].x1; _atArrow[5].y1 = _atArrow[1].y1 + 1;
    _atArrow[5].x2 = _atArrow[1].x2; _atArrow[5].y2 = _atArrow[1].y2 + 1;
    // Down button
    n32X1 = n32X + 1;                  n32X2 = n32X1 + 13 - 1;
    n32Y2 = n32Y + un32Height - 1 - 1; n32Y1 = n32Y2 - (un32Height / 2) + 2;
    _tDn.x1     = n32X1;     _tDn.y1     = n32Y1;     _tDn.x2     = n32X2; _tDn.y2     = n32Y2;
    _atDn[0].x1 = n32X1;     _atDn[0].y1 = n32Y1;     _atDn[0].x2 = n32X1; _atDn[0].y2 = n32Y2;
    _atDn[1].x1 = n32X1 + 1; _atDn[1].y1 = n32Y1;     _atDn[1].x2 = n32X2; _atDn[1].y2 = n32Y1;
    _atDn[2].x1 = n32X1 + 1; _atDn[2].y1 = n32Y2;     _atDn[2].x2 = n32X2; _atDn[2].y2 = n32Y2;
    _atDn[3].x1 = n32X2;     _atDn[3].y1 = n32Y1 + 1; _atDn[3].x2 = n32X2; _atDn[3].y2 = n32Y2 - 1;
    n32Xc = n32X1 + 6; n32Yc = n32Y1 + (un32Height / 4) - 1;
    _atArrow[2].x1 = n32Xc - 4; _atArrow[2].x2 = n32Xc;     _atArrow[2].y1 = n32Yc - 2; _atArrow[2].y2 = n32Yc + 2;
    _atArrow[3].x1 = n32Xc;     _atArrow[3].x2 = n32Xc + 4; _atArrow[3].y1 = n32Yc + 2; _atArrow[3].y2 = n32Yc - 2;
    _atArrow[6].x1 = _atArrow[2].x1; _atArrow[6].y1 = _atArrow[2].y1 + 1;
    _atArrow[6].x2 = _atArrow[2].x2; _atArrow[6].y2 = _atArrow[2].y2 + 1;
    _atArrow[7].x1 = _atArrow[3].x1; _atArrow[7].y1 = _atArrow[3].y1 + 1;
    _atArrow[7].x2 = _atArrow[3].x2; _atArrow[7].y2 = _atArrow[3].y2 + 1;
  }
  else
    VError(VERR_NotImplemented, _NFL_, "This type of spin button is not implemented yet");

  // Calculate text coordinates
  if(_eTA == EHTA_Left)
    _n32Xo = _atText[1].x1 + 5;
  else if(_eTA == EHTA_Center)
    _n32Xo = _atText[1].x1 + ((_atText[1].x2 - _atText[1].x1) / 2);
  else
    _n32Xo = _atText[1].x2 - 5;
  _n32Xt = _n32Xo;

  // Initialize the default value
  if(eValueType == ESBV_Integer) {
    SetValueProp(0, 100, 1);
    SetValue(0);
  }
  else {
    SetValuePropF(0.0, 10.0, 0.1);
    SetValueF(0.0);
  }
  _bDirty = true;
}

void CSpinButton::UpdateValue() const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Update value
  if(_eVT == ESBV_Integer) {
    if(_bWrap) {
      if(_uVP.tI.n32Cur < _uVP.tI.n32Min) _uVP.tI.n32Cur = _uVP.tI.n32Max;
      if(_uVP.tI.n32Cur > _uVP.tI.n32Max) _uVP.tI.n32Cur = _uVP.tI.n32Min;
    }
    else {
      if(_uVP.tI.n32Cur < _uVP.tI.n32Min) _uVP.tI.n32Cur = _uVP.tI.n32Min;
      if(_uVP.tI.n32Cur > _uVP.tI.n32Max) _uVP.tI.n32Cur = _uVP.tI.n32Max;
    }
  }
  else {
    if(_bWrap) {
      if(_uVP.tF.fCur < _uVP.tF.fMin) _uVP.tF.fCur = _uVP.tF.fMax;
      if(_uVP.tF.fCur > _uVP.tF.fMax) _uVP.tF.fCur = _uVP.tF.fMin;
    }
    else {
      if(_uVP.tF.fCur < _uVP.tF.fMin) _uVP.tF.fCur = _uVP.tF.fMin;
      if(_uVP.tF.fCur > _uVP.tF.fMax) _uVP.tF.fCur = _uVP.tF.fMax;
    }
  }

  // Update string buffer
  if(_eVT == ESBV_Integer)     snprintf(_acValue, sizeof(_acValue), "%d",   _uVP.tI.n32Cur);
  else if(_eVT == ESBV_Float1) snprintf(_acValue, sizeof(_acValue), "%.1f", _uVP.tF.fCur);
  else if(_eVT == ESBV_Float2) snprintf(_acValue, sizeof(_acValue), "%.2f", _uVP.tF.fCur);
  else if(_eVT == ESBV_Float3) snprintf(_acValue, sizeof(_acValue), "%.3f", _uVP.tF.fCur);
  else if(_eVT == ESBV_Float4) snprintf(_acValue, sizeof(_acValue), "%.4f", _uVP.tF.fCur);
  else                         snprintf(_acValue, sizeof(_acValue), "%.5f", _uVP.tF.fCur);
  _un8Len = strlen(_acValue);
  for(size_t i = 0; i < _un8Len; ++i) {
    if(_acValue[i] == '.') {
      _acValue[i] = CBasicAppObjects::DecimalSeparator();
      break;
    }
  }

  // Calculate coordinates
  if(_eTA == EHTA_Left)
    _n32Xt = _n32Xo;
  else if(_eTA == EHTA_Center)
    _n32Xt = _n32Xo -(XTextWidth(*vidEnv._tX11Style.ptFont, _acValue, _un8Len) / 2);
  else
    _n32Xt = _n32Xo - XTextWidth(*vidEnv._tX11Style.ptFont, _acValue, _un8Len);
}

bool CSpinButton::ToggleValue(int32_t n32MouseX, int32_t n32MouseY)
{
  // Check
  if(!_tBD.bEnabled) return(false);

  // Check up buttons
  if(MouseOver(_tUp.x1, _tUp.y1, _tUp.x2, _tUp.y2, n32MouseX, n32MouseY)) {
    if(_eVT == ESBV_Integer)
      _uVP.tI.n32Cur += _uVP.tI.n32Step;
    else
      _uVP.tF.fCur += _uVP.tF.fStep;
    return(true);
  }
  // Check down buttons
  else if(MouseOver(_tDn.x1, _tDn.y1, _tDn.x2, _tDn.y2, n32MouseX, n32MouseY)) {
    if(_eVT == ESBV_Integer)
      _uVP.tI.n32Cur -= _uVP.tI.n32Step;
    else
      _uVP.tF.fCur -= _uVP.tF.fStep;
    return(true);
  }
  return(false);
}

void CSpinButton::DrawControl() const
{
  DrawControl(false, false);
}

void CSpinButton::DrawControl(bool bUpPressed, bool bDnPressed) const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw outline
  vidEnv.SetFillStyle(FILL_SOLID);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
  XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                 _tOutline.x, _tOutline.y, _tOutline.width, _tOutline.height);
  if(vidEnv._tX11Style.xFillStyle == FILL_GRADIENT) {
    //! \bug May not work as expected if more spin button styles got implemented later (this should be corrected later)
    vidEnv.DrawGradientBox(_tUp.x1, _tUp.y1, _tUp.x2 - _tUp.x1, _tDn.y2 - _tUp.y1,
                               vidEnv._tX11Style.xLightBorder,
                               vidEnv._tX11Style.xDarkBorder,
                               vidEnv._tX11Style.xFace);
  }

  vidEnv.SetForegroundColor(vidEnv._tX11Style.xOutline);
  XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
    _tOutline.x, _tOutline.y, _tOutline.width, _tOutline.height);

  // Draw buttons
  if(_tBD.bEnabled) {
    // Draw arrows
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
    XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  const_cast<XSegment*>(_atArrow), 8);
    // Draw up button
    if(bUpPressed) {
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
      XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                    const_cast<XSegment*>(_atUp), 2);
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
      XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                    const_cast<XSegment*>(&_atUp[2]), 2);
    }
    else {
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
      XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                    const_cast<XSegment*>( _atUp), 2);
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
      XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                    const_cast<XSegment*>(&_atUp[2]), 2);
    }
    // Draw down button
    if(bDnPressed) {
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
      XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                    const_cast<XSegment*>(_atDn), 2);
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
      XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                    const_cast<XSegment*>(&_atDn[2]), 2);
    }
    else {
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
      XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                    const_cast<XSegment*>(_atDn), 2);
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
      XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                    const_cast<XSegment*>(&_atDn[2]), 2);
    }
  }
  else {
    // Draw arrows
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  const_cast<XSegment*>(&_atArrow[4]), 4);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
    XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  const_cast<XSegment*>(_atArrow), 4);
    // Draw up button
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  const_cast<XSegment*>(_atUp), 2);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
    XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  const_cast<XSegment*>(&_atUp[2]), 2);
    // Draw down button
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  const_cast<XSegment*>(_atDn), 2);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
    XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  const_cast<XSegment*>(&_atDn[2]), 2);
  }

  // Draw text
  if(_bDirty) {
    _bDirty = false; 
    UpdateValue();
  }
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                const_cast<XSegment*>(_atText), 2);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                const_cast<XSegment*>(&_atText[2]), 2);
  vidEnv.SetFont(vidEnv._tX11Style.ptFont);
  if(_tBD.bEnabled) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt, _n32Yt, _acValue,  _un8Len);
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt, _n32Yt, _acValue,  _un8Len);
#endif
  }
  else {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt, _n32Yt, _acValue,  _un8Len);
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt, _n32Yt, _acValue,  _un8Len);
#endif
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt-1, _n32Yt-1, _acValue,  _un8Len);
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt-1, _n32Yt-1, _acValue,  _un8Len);
#endif
  }
}

void CSpinButton::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  if(MouseOver(_tUp.x1, _tUp.y1, _tUp.x2, _tUp.y2, n32MouseX, n32MouseY))
    DrawControl(bPressed, false);
  else if(MouseOver(_tDn.x1, _tDn.y1, _tDn.x2, _tDn.y2, n32MouseX, n32MouseY))
    DrawControl(false, bPressed);
  else
    DrawControl(false, false);
}


//
// Extended-spin-button control
//
//! \todo WRITE ME !!!


//
// Slider-bar control
//
//! \todo WRITE ME !!!

} // namespace vge
